import numpy as np

from asr.core import command


# ---------- Recipe tests ---------- #


# Add test(s)                                                               XXX


# ---------- Webpanel ---------- #


def webpanel(row, key_descriptions):
    from asr.database.browser import fig

    panel = {'title': 'Electronic band structure and projected DOS (PBE)',
             'columns': [[fig('pbe-projected-bs.png', link='empty')], []],
             'plot_descriptions': [{'function': projected_bs_pbe,
                                    'filenames': ['pbe-projected-bs.png']}],
             'sort': 14}

    return [panel]


# ---------- Main functionality ---------- #


@command(module='asr.projected_bandstructure',
         requires=['results-asr.gs.json', 'bs.gpw',
                   'results-asr.bandstructure.json'],
         dependencies=['asr.gs', 'asr.bandstructure'],
         webpanel=webpanel)
def main():
    from gpaw import GPAW

    # Get bandstructure calculation
    calc = GPAW('bs.gpw', txt=None)

    results = {}

    # Extract projections
    weight_skni, yl_i = get_orbital_ldos(calc)
    results['weight_skni'] = weight_skni
    results['yl_i'] = yl_i
    results['symbols'] = calc.atoms.get_chemical_symbols()
    return results


# ---------- Recipe methodology ---------- #


def get_orbital_ldos(calc):
    """Get the projection weights on different orbitals.

    Returns
    -------
    weight_skni : nd.array
        weight of each projector (indexed by (s, k, n)) on orbitals i
    yl_i : list
        symbol and orbital angular momentum string ('y,l') of each orbital i
    """
    from ase.utils import DevNull
    from ase.parallel import parprint
    import gpaw.mpi as mpi
    from gpaw.utilities.dos import raw_orbital_LDOS
    from gpaw.utilities.progressbar import ProgressBar
    from asr.pdos import get_l_a

    ns = calc.get_number_of_spins()
    zs = calc.atoms.get_atomic_numbers()
    chem_symbols = calc.atoms.get_chemical_symbols()
    l_a = get_l_a(zs)

    # We distinguish in (chemical symbol(y), angular momentum (l)),
    # that is if there are multiple atoms in the unit cell of the same chemical
    # species, their weights are added together.
    # x index for each unique atom
    a_x = [a for a in l_a for l in l_a[a]]
    l_x = [l for a in l_a for l in l_a[a]]
    # Get i index for each unique symbol
    yl_i = []
    i_x = []
    for a, l in zip(a_x, l_x):
        symbol = chem_symbols[a]
        yl = ','.join([str(symbol), str(l)])
        if yl in yl_i:
            i = yl_i.index(yl)
        else:
            i = len(yl_i)
            yl_i.append(yl)
        i_x.append(i)

    # Allocate output array
    nk, nb = len(calc.get_ibz_k_points()), calc.get_number_of_bands()
    weight_skni = np.zeros((ns, nk, nb, len(yl_i)))

    # Set up progressbar
    ali_x = [(a, l, i) for (a, l, i) in zip(a_x, l_x, i_x)]
    parprint('Computing orbital ldos')
    if mpi.world.rank == 0:
        pb = ProgressBar()
    else:
        devnull = DevNull()
        pb = ProgressBar(devnull)

    for _, (a, l, i) in pb.enumerate(ali_x):
        # Extract weights
        for s in range(ns):
            __, weights = raw_orbital_LDOS(calc, a, s, l)
            weight_kn = weights.reshape((nk, nb))
            # Renormalize (don't include reciprocal space volume in weight)
            weight_kn /= calc.wfs.kd.weight_k[:, np.newaxis]
            weight_skni[s, :, :, i] += weight_kn

    return weight_skni, yl_i


# ---------- Plotting ---------- #


def get_yl_ordering(yl_i, symbols):
    """Get standardized yl ordering of keys.

    Parameters
    ----------
    yl_i : list
        see get_orbital_ldos
    symbols : list
        Sort symbols after index in this list

    Returns
    -------
    c_i : list
        ordered index for each i
    """
    # Setup sili (symbol index, angular momentum index) key
    def sili(yl):
        y, L = yl.split(',')
        # Symbols list can have multiple entries of the same symbol
        # ex. ['O', 'Fe', 'O']. In this case 'O' will have index 0 and
        # 'Fe' will have index 1.
        si = symbols.index(y)
        li = ['s', 'p', 'd', 'f'].index(L)
        return f'{si}{li}'

    i_c = [iyl[0] for iyl in sorted(enumerate(yl_i), key=lambda t: sili(t[1]))]
    return [i_c.index(i) for i in range(len(yl_i))]


def get_bs_sampling(bsp, npoints=40):
    """Sample band structure as evenly as possible.

    Allways include special points.

    Parameters
    ----------
    bsp : obj
        ase.dft.band_structure.BandStructurePlot object
    npoints : int
        number of k-points to sample along band structure

    Returns
    -------
    chosenx_x : 1d np.array
        chosen band structure coordinates
    k_x : 1d np.array
        chosen k-point indices
    """
    # Get band structure coordinates and unique labels
    xcoords, label_xcoords, orig_labels = bsp.bs.get_labels()
    label_xcoords = np.unique(label_xcoords)

    # Reserve one point for each special point
    nonspoints = npoints - len(label_xcoords)
    assert nonspoints >= 0
    assert npoints <= len(xcoords)

    # Slice xcoords into seperate subpaths
    xcoords_lx = []
    subpl_l = []
    lastx = 0.
    for labelx in label_xcoords:
        xcoords_x = xcoords[np.logical_and(xcoords >= lastx,
                                           xcoords <= labelx)]
        xcoords_lx.append(xcoords_x)
        subpl_l.append(xcoords_x[-1] - xcoords_x[0])  # Length of subpath
        lastx = labelx

    # Distribute trivial k-points based on length of slices
    pathlength = sum(subpl_l)
    unitlength = pathlength / (nonspoints + 1)
    # Floor npoints and length remainder for each subpath
    subpnp_l, subprl_l = np.divmod(subpl_l, unitlength)
    subpnp_l = subpnp_l.astype(int)
    # Distribute remainders
    points_left = nonspoints - np.sum(subpnp_l)
    subpnp_l[np.argsort(subprl_l)[-points_left:]] += 1

    # Choose points on each sub path
    chosenx_x = []
    for subpnp, xcoords_x in zip(subpnp_l, xcoords_lx):
        # Evenly spaced indices
        x_p = np.unique(np.round(np.linspace(0, len(xcoords_x) - 1,
                                             subpnp + 2)).astype(int))
        chosenx_x += list(xcoords_x[x_p][:-1])  # each subpath includes start
    chosenx_x.append(xcoords[-1])  # Add end of path

    # Get k-indeces
    chosenx_x = np.array(chosenx_x)
    x_y, k_y = np.where(chosenx_x[:, np.newaxis] == xcoords[np.newaxis, :])
    x_x, y_x = np.unique(x_y, return_index=True)
    k_x = k_y[y_x]

    return chosenx_x, k_x


def get_pie_slice(theta0, theta, s=36., res=126):
    """Get a single pie slice marker.

    Parameters
    ----------
    theta0 : float
        angle in which to start slice
    theta : float
        angle that pie slice should cover
    s : float
        marker size
    res : int
        resolution of pie (in points around the circumference)

    Returns
    -------
    pie : matplotlib.pyplot.scatter option dictionary
    """
    assert -np.pi / res <= theta0 and theta0 <= 2. * np.pi + np.pi / res
    assert -np.pi / res <= theta and theta <= 2. * np.pi + np.pi / res

    angles = np.linspace(theta0, theta0 + theta,
                         int(np.ceil(res * theta / (2 * np.pi))))
    x = [0] + np.cos(angles).tolist()
    y = [0] + np.sin(angles).tolist()
    xy = np.column_stack([x, y])
    size = s * np.abs(xy).max() ** 2

    return {'marker': xy, 's': size, 'linewidths': 0.0}


def get_pie_markers(weight_xi, scale_marker=True, s=36., res=126):
    """Get pie markers corresponding to a 2D array of weights.

    Parameters
    ----------
    weight_xi : 2d np.array
    scale_marker : bool
        using sum of weights as scale for markersize
    s, res : see get_pie_slice

    Returns
    -------
    pie_xi : list of lists of mpl option dictionaries
    """
    assert np.all(weight_xi >= 0.)

    pie_xi = []
    for weight_i in weight_xi:
        pie_i = []
        # Normalize by total weight
        totweight = np.sum(weight_i)
        r0 = 0.
        for weight in weight_i:
            # Weight fraction
            r1 = weight / totweight

            # Get slice
            pie = get_pie_slice(2 * np.pi * r0,
                                2 * np.pi * r1, s=s, res=res)
            if scale_marker:
                pie['s'] *= totweight

            pie_i.append(pie)
            r0 += r1
        pie_xi.append(pie_i)

    return pie_xi


def projected_bs_pbe(row,
                     filename='pbe-projected-bs.png',
                     figsize=(5.5, 5),
                     fontsize=10):  # Choose input parameters               XXX
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    from matplotlib.lines import Line2D
    # import pylab
    import numpy as np
    from ase.dft.band_structure import BandStructure, BandStructurePlot
    mpl.rcParams['font.size'] = fontsize

    # Extract projections data
    data = row.data.get('results-asr.projected_bandstructure.json')
    weight_skni = data['weight_skni']
    yl_i = data['yl_i']

    # Get color indeces
    c_i = get_yl_ordering(yl_i, data['symbols'])

    # Extract band structure data
    d = row.data.get('results-asr.bandstructure.json')
    path = d['bs_nosoc']['path']
    ef = d['bs_nosoc']['efermi']

    # If a vacuum energy is available, use it as a reference
    ref = row.get('evac', d.get('bs_nosoc').get('efermi'))
    if row.get('evac') is not None:
        label = r'$E - E_\mathrm{vac}$ [eV]'
    else:
        label = r'$E - E_\mathrm{F}$ [eV]'

    # Determine plotting window based on band gap
    gaps = row.data.get('results-asr.gs.json', {}).get('gaps_nosoc', {})
    if gaps.get('vbm'):
        emin = gaps.get('vbm') - 3
    else:
        emin = ef - 3
    if gaps.get('cbm'):
        emax = gaps.get('cbm') + 3
    else:
        emax = ef + 3

    # Take bands with energies in range
    e_skn = d['bs_nosoc']['energies']
    inrange_skn = np.logical_and(e_skn > emin, e_skn < emax)
    inrange_n = np.any(np.any(inrange_skn, axis=1), axis=0)
    e_skn = e_skn[:, :, inrange_n]
    weight_skni = weight_skni[:, :, inrange_n, :]

    # Use band structure objects to plot outline
    bs = BandStructure(path, e_skn - ref, ef - ref)
    # Use colors if spin-polarized
    if e_skn.shape[0] == 2:
        spincolors = ['0.8', '0.4']
    else:
        spincolors = ['0.8'] * e_skn.shape[0]
    style = dict(
        colors=spincolors,
        ls='-',
        lw=1.0,
        zorder=0)
    ax = plt.figure(figsize=figsize).add_subplot(111)
    bsp = BandStructurePlot(bs)
    bsp.plot(ax=ax, show=False, emin=emin - ref, emax=emax - ref,
             ylabel=label, **style)

    npoints = 40  # input variable?
    xcoords, k_x = get_bs_sampling(bsp, npoints=npoints)

    # Generate energy and weight arrays based on band structure sampling
    ns, nk, nb = e_skn.shape
    s_u = np.array([s for s in range(ns) for n in range(nb)])
    n_u = np.array([n for s in range(ns) for n in range(nb)])
    e_ux = e_skn[s_u[:, np.newaxis],
                 k_x[np.newaxis, :],
                 n_u[:, np.newaxis]] - ref
    weight_uxi = weight_skni[s_u[:, np.newaxis],
                             k_x[np.newaxis, :],
                             n_u[:, np.newaxis], :]
    # Plot projections
    # Choose some plotting format                                           XXX
    markersize = 36.
    res = 64  # input variable?                                             XXX
    for e_x, weight_xi in zip(e_ux, weight_uxi):

        # Weights as pie chart
        pie_xi = get_pie_markers(weight_xi, s=markersize,
                                 scale_marker=False, res=res)
        for x, e, weight_i, pie_i in zip(xcoords, e_x, weight_xi, pie_xi):
            # totweight = np.sum(weight_i)
            for i, pie in enumerate(pie_i):
                ax.scatter(x, e, facecolor='C{}'.format(c_i[i]),
                           zorder=3, **pie)
            # Plot radius = 1 circle for comparison of total weights
            # cx = np.cos(np.linspace(0., 2 * np.pi, res)).tolist()
            # cy = np.sin(np.linspace(0., 2 * np.pi, res)).tolist()
            # xy = np.column_stack([cx, cy])
            # ax.scatter(x, e, marker=xy, linewidth=.5,
            #            s=markersize * np.abs(xy).max() ** 2,
            #            facecolor='none', edgecolor='black', zorder=4)

        # Marker size depending on each weight
        # for x, e, weight_i in zip(xcoords, e_x, weight_xi):
        #     # Sort orbital after weight
        #     i_si = np.argsort(weight_i)
        #     # Calculate accumulated weight and use it for area of marker.
        #     # Join orbitals with less than 10% weight in a neutral marker
        #     totweight = np.sum(weight_i)
        #     accweight = 0.
        #     neutralweight = 0.
        #     c_pi = []
        #     a_pi = []
        #     for i, weight in zip(i_si, weight_i[i_si]):
        #         if weight < 0.1 * totweight:
        #             neutralweight += weight
        #         else:
        #             if neutralweight is not None:
        #                 # Done with small weights, store neutral marker
        #                 a_pi.append(markersize * neutralweight)
        #                 c_pi.append('xkcd:grey')
        #                 accweight += neutralweight
        #                 neutralweight = None
        #             # Add orbital as is
        #             accweight += weight
        #             a_pi.append(markersize * accweight)
        #             c_pi.append('C{}'.format(c_i[i]))
        #     # Plot points from largest to smallest
        #     for a, c in zip(reversed(a_pi), reversed(c_pi)):
        #         ax.scatter(x, e, color=c, s=a, zorder=3)
        # # Plot radius = 1 circle for comparison of total weights
        # ax.scatter(xcoords, e_x, marker='o', s=markersize, linewidth=.5,
        #            facecolor='none', edgecolor='black', zorder=4)

        # Marker color depending on largest weight
        # c_x = [c_i[i] for i in np.argmax(weight_xi, axis=1)]
        # for x, e, c in zip(xcoords, e_x, c_x):
        #     ax.scatter(x, e, color='C{}'.format(c), s=markersize, zorder=3)

    # Set legend
    # Get "pac-man" style pie slice marker
    pie = get_pie_slice(1. * np.pi / 4.,
                        3. * np.pi / 2., s=markersize, res=res)
    # Generate markers for legend
    legend_markers = []
    for i, yl in enumerate(yl_i):
        legend_markers.append(Line2D([0], [0],
                                     mfc='C{}'.format(c_i[i]), mew=0.0,
                                     marker=pie['marker'], ms=3. * np.pi,
                                     linewidth=0.0))
    # Generate legend
    plt.legend(legend_markers, [yl.replace(',', ' (') + ')' for yl in yl_i],
               title='Fraction of orbital projection weights:',
               bbox_to_anchor=(0., -0.08, 1., 0.), loc='upper left',
               ncol=3, mode="expand", borderaxespad=0.)

    # ax.figure.set_figheight(1.2 * ax.figure.get_figheight())
    plt.savefig(filename, bbox_inches='tight')


if __name__ == '__main__':
    main.cli()



from asr.core import command


def bz_vertices(cell):
    import numpy as np
    from scipy.spatial import Voronoi
    icell = np.linalg.inv(cell) * 2 * np.pi
    ind = np.indices((3, 3)).reshape((2, 9)) - 1
    G = np.dot(icell, ind).T
    vor = Voronoi(G)
    bz1 = []
    for vertices, points in zip(vor.ridge_vertices, vor.ridge_points):
        if -1 not in vertices and 4 in points:
            normal = G[points].sum(0)
            normal /= (normal**2).sum()**0.5
            bz1.append((vor.vertices[vertices], normal))
    return bz1


def find_contours(eigs_nk, bzk_kv, s_nk=None):
    import numpy as np
    from scipy.interpolate import griddata
    import matplotlib.pyplot as plt
    minx = np.min(bzk_kv[:, 0:2])
    maxx = np.max(bzk_kv[:, 0:2])

    npoints = 1000

    xi = np.linspace(minx, maxx, npoints)
    yi = np.linspace(minx, maxx, npoints)

    zis = []
    for eigs_k in eigs_nk:
        zi = griddata((bzk_kv[:, 0], bzk_kv[:, 1]), eigs_k,
                      (xi[None, :], yi[:, None]), method='cubic')
        zis.append(zi)

    contours = []
    for n, zi in enumerate(zis):
        cs = plt.contour(xi, yi, zi, levels=[0])
        paths = cs.collections[0].get_paths()

        for path in paths:
            vertices = []
            for vertex in path.iter_segments(simplify=False):
                vertices.append(np.array((vertex[0][0],
                                          vertex[0][1],
                                          vertex[1], 0), float))
            vertices = np.array(vertices)
            if s_nk is not None:
                si = griddata((bzk_kv[:, 0], bzk_kv[:, 1]), s_nk[n],
                              vertices[:, :2], method='cubic')
                vertices[:, -1] = si
            contours.append(vertices)

    return contours


def webpanel(row, key_descriptions):
    from asr.database.browser import fig

    panel = {'title': 'Fermi surface',
             'columns': [[fig('fermi_surface.png')]],
             'plot_descriptions': [{'function': plot_fermi,
                                    'filenames': ['fermi_surface.png']}],
             'sort': 13}

    return [panel]


def plot_fermi(row, fname,
               annotate=True, fontsize=10, svbm=100, scbm=40, lwvbm=2.5,
               sfs=1, dpi=200, scale=None, scalecb=None,
               bbox_to_anchor=None, angle=0):
    from ase.geometry.cell import Cell
    from matplotlib import pyplot as plt
    cell = Cell(row.cell)
    lat = cell.get_bravais_lattice(pbc=row.pbc)
    plt.figure(figsize=(4, 3))
    ax = lat.plot_bz(vectors=False)
    add_fermi(row, ax=ax, annotate=annotate, s=sfs, scale=scalecb)
    plt.savefig(fname, dpi=dpi)
    plt.close()


def add_fermi(row, ax, annotate=True, s=0.25, scale=None, angle=0,):
    from matplotlib import pyplot as plt
    import matplotlib.colors as colors
    import numpy as np
    verts = row.data['results-asr.fermisurface.json']['contours'].copy()
    normalize = colors.Normalize(vmin=-1, vmax=1)
    rotate = np.array([[np.cos(angle), -np.sin(angle), 0],
                       [np.sin(angle), np.cos(angle), 0],
                       [0, 0, 1]])
    verts[:, :2] /= (2 * np.pi)
    verts[:, :2] = np.dot(rotate[:2, :2], verts[:, :2].T).T
    im = ax.scatter(verts[:, 0], verts[:, 1], c=verts[:, -1],
                    s=s, cmap='viridis', marker=',',
                    norm=normalize, alpha=1, zorder=2)
    rect = np.array([0.85, 0.2, 0.025, 0.6])
    if scale is not None:
        center = np.array([0, 0.5, 0, 0])
        rect = (rect - center) * scale + center
    cbaxes = plt.gcf().add_axes(rect)
    cbar = plt.colorbar(im, cax=cbaxes, ticks=[-1, -0.5, 0, 0.5, 1])
    cbar.ax.tick_params()
    cbar.set_label('$\\langle S_z \\rangle$')
    # ax.annotate('Fermi surface', xy=(0.5, 1), ha='center',
    #             va='top', xycoords='axes fraction')

    return cbaxes


@command('asr.fermisurface',
         webpanel=webpanel,
         requires=['gs.gpw', 'results-asr.structureinfo.json'],
         dependencies=['asr.gs@calculate', 'asr.structureinfo'])
def main():
    import numpy as np
    from gpaw import GPAW
    from asr.utils.gpw2eigs import gpw2eigs
    from gpaw.kpt_descriptor import to1bz
    from asr.magnetic_anisotropy import get_spin_axis, get_spin_index
    theta, phi = get_spin_axis()
    eigs_km, ef, s_kvm = gpw2eigs('gs.gpw', return_spin=True,
                                  theta=theta, phi=phi)
    eigs_mk = eigs_km.T
    eigs_mk -= ef
    calc = GPAW('gs.gpw', txt=None)
    s_mk = s_kvm[:, get_spin_index()].T

    A_cv = calc.atoms.get_cell()
    B_cv = np.linalg.inv(A_cv).T * 2 * np.pi

    bzk_kc = calc.wfs.kd.bzk_kc
    bzk_kv = np.dot(bzk_kc, B_cv)

    contours = []
    selection = ~np.logical_or(eigs_mk.max(1) < 0, eigs_mk.min(1) > 0)
    eigs_mk = eigs_mk[selection, :]
    s_mk = s_mk[selection, :]
    bz2ibz_k = calc.wfs.kd.bz2ibz_k
    eigs_mk = eigs_mk[:, bz2ibz_k]
    s_mk = s_mk[:, bz2ibz_k]

    n = 5
    N_xc = np.indices((n, n, 1)).reshape((3, n**2)).T - n // 2
    N_xc += np.array((0, 0, n // 2))
    N_xv = np.dot(N_xc, B_cv)

    eigs_mk = np.repeat(eigs_mk, n**2, axis=1)
    s_mk = np.repeat(s_mk, n**2, axis=1)
    tmpbzk_kv = (bzk_kv[:, np.newaxis] + N_xv[np.newaxis]).reshape(-1, 3)
    tmpcontours = find_contours(eigs_mk, tmpbzk_kv, s_nk=s_mk)

    # Only include contours with a part in 1st BZ
    for cnt in tmpcontours:
        k_kv = cnt.copy()
        k_kv = k_kv[:, :3]
        k_kv[:, 2] = 0
        k_kc = np.dot(k_kv, A_cv.T) / (2 * np.pi)
        inds_k = np.linalg.norm(to1bz(k_kc, A_cv) - k_kc, axis=1) < 1e-8
        if (inds_k).any():
            contours.append(cnt[inds_k, :])

    contours = np.concatenate(contours)
    data = {'contours': contours}
    return data


if __name__ == '__main__':
    main.cli()



from pathlib import Path
from asr.core import command, option
import click


@command('asr.setup.defects',
         creates=['general_parameters.json'])
@option('-a', '--atomfile', type=str,
        help='Atomic structure.')
@option('-q', '--chargestates', type=int,
        help='Charge states included (-q, ..., +q).')
@option('--supercell', nargs=3, type=click.Tuple([int, int, int]),
        help='List of repetitions in lat. vector directions [N_x, N_y, N_z]')
@option('--maxsize', type=float,
        help='Maximum supercell size in Å.')
@option('--intrinsic', type=bool,
        help='Specify whether you want to incorporate anti-site defects.')
@option('--vacancies', type=bool,
        help='Specify whether you want to incorporate vacancies.')
def main(atomfile='unrelaxed.json', chargestates=3, supercell=[0, 0, 0],
         maxsize=8, intrinsic=True, vacancies=True):
    """Set up defect structures for a given host.

    Recipe setting up all possible defects within a reasonable supercell as
    well as the respective pristine system for a given input structure.
    Defects include: vacancies, anti-site defects. For a given primitive input
    structure this recipe will create a directory tree in the following way:

    For the example of MoS2:

    - There has to be a 'unrelaxed.json' file with the primitive structure
      of the desired system in the folder you run setup.defects. The tree
      structure will then look like this

    .                                                                        '
    ├── general_parameters.json                                              '
    ├── MoS2_231.v_at_b0                                                     '
    │   ├── charge_0                                                         '
    │   │   ├── params.json                                                  '
    │   │   └── unrelaxed.json                                               '
    │   ├── charge_1                                                         '
    │   │   ├── ...                                                          '
    │   │   .                                                                '
    │   .                                                                    '
    │                                                                        '
    ├── MoS2_231.Mo_at_i1                                                    '
    │   ├── charge_0                                                         '
    .   .                                                                    '
    .                                                                        '
    ├── pristine                                                             '
    │   └── neutral                                                          '
    │       ├── params.json                                                  '
    │       └── unrelaxed.json                                               '
    ├── pristine_sc                                                          '
    │   └── neutral                                                          '
    │       ├── params.json                                                  '
    │       └── unrelaxed.json                                               '
    ├── results_setup.defects.json                                           '
    └── unrelaxed.json                                                       '

    - Here, the notation for the defects is the following:
      'formula_supercellsize.defect_at_substitutionposition' where 'v'
      denotes a vacancy
    - In the resulting folders you can find the unrelaxed structures, as
      well as a 'params.json' file which contains the charge states of the
      different defect structures.

    """
    from ase.io import read
    import numpy as np

    # first, read input atomic structure and store it in ase's atoms object
    structure = read(atomfile)
    print('INFO: starting recipe for setting up defect systems of '
          '{} host system.'.format(structure.symbols))

    # check dimensionality of initial parent structure
    nd = int(np.sum(structure.get_pbc()))
    if nd == 3:
        is2d = False
    elif nd == 2:
        is2d = True
    elif nd == 1:
        raise NotImplementedError('Setup defects not implemented for 1D '
                                  f'structures')
    # set up the different defect systems and store their properties
    # in a dictionary
    structure_dict = setup_defects(structure=structure, intrinsic=intrinsic,
                                   charge_states=chargestates,
                                   vacancies=vacancies, sc=supercell,
                                   max_lattice=maxsize, is_2D=is2d)

    # based on this dictionary, create a folder structure for all defects
    # and respective charge states
    create_folder_structure(structure, structure_dict, chargestates,
                            intrinsic=intrinsic, vacancies=vacancies,
                            sc=supercell, max_lattice=maxsize, is_2D=is2d)

    return None


def setup_supercell(structure, max_lattice, is_2D):
    """Set up the supercell of a given structure depending.

    Set up the supercell of a given structure depending on a
    maximum supercell lattice vector length for 2D or 3D structures.

    Parameters
    ----------
    structure
        input structure (primitive cell)
    max_lattice : float
        maximum supercell lattice vector length in Å
    is_2D : bool
        choose 2D or 3D supercell (is_2D=False)

    Returns
    -------
    structure_sc
        supercell structure
    """
    for x in range(1, 50):
        struc_temp = structure.repeat((x, 1, 1))
        diff = struc_temp.get_distance(0, -1)
        if diff > max_lattice:
            x_size = x - 1
            break
    for y in range(1, 50):
        struc_temp = structure.repeat((1, y, 1))
        diff = struc_temp.get_distance(0, -1)
        if diff > max_lattice:
            y_size = y - 1
            break
    if not is_2D:
        for z in range(1, 50):
            struc_temp = structure.repeat((1, 1, z))
            diff = struc_temp.get_distance(0, -1)
            if diff > max_lattice:
                z_size = z - 1
                break
    else:
        z_size = 1

    print('INFO: setting up supercell: ({0}, {1}, {2})'.format(
          x_size, y_size, z_size))
    x_size = max(1, x_size)
    y_size = max(1, y_size)
    z_size = max(1, z_size)
    structure_sc = structure.repeat((x_size, y_size, z_size))

    return structure_sc, x_size, y_size, z_size


def setup_defects(structure, intrinsic, charge_states, vacancies, sc,
                  max_lattice, is_2D):
    """Set up all possible defects.

    Sets up all possible defects (i.e. vacancies, intrinsic anti-sites,
    extrinsic point defects('extrinsic=True')) for a given structure.

    Parameters
    ----------
    structure
        input structure (primitive cell)
    intrinsic : bool
        incorporate intrinsic point defects
    vacancies : bool
        incorporate vacancies

    Returns
    -------
    structure_dict : dict
        dictionary of all possible defect configurations
        of the given structure with different charge
        states. The dictionary is built up in the
        following way: see folder structure in 'main()'.
    """
    import spglib

    # set up artificial array in order to check for equivalent positions later
    cell = (structure.cell.array, structure.get_scaled_positions(),
            structure.numbers)

    # set up a dictionary
    structure_dict = {}
    formula = structure.symbols

    # first, find the desired supercell
    if sc == (0, 0, 0):
        pristine, N_x, N_y, N_z = setup_supercell(
            structure, max_lattice, is_2D)
    else:
        N_x = sc[0]
        N_y = sc[1]
        N_z = sc[2]
        print('INFO: setting up supercell: ({0}, {1}, {2})'.format(
              N_x, N_y, N_z))
        pristine = structure.repeat((N_x, N_y, N_z))
    parameters = {}
    string = 'defects.pristine_sc'
    parameters['asr.relax'] = {}
    parameters['asr.gs@calculate'] = {}
    structure_dict[string] = {'structure': pristine, 'parameters': parameters}

    # incorporate the possible vacancies
    dataset = spglib.get_symmetry_dataset(cell)
    eq_pos = dataset.get('equivalent_atoms')
    wyckoffs = dataset.get('wyckoffs')

    finished_list = []
    if vacancies:
        temp_dict = {}
        for i in range(len(structure)):
            if not eq_pos[i] in finished_list:
                vacancy = pristine.copy()
                vacancy.pop(i)
                string = 'defects.{0}_{1}{2}{3}.v_{4}{5}'.format(
                         formula, N_x, N_y, N_z, wyckoffs[i], i)
                charge_dict = {}
                for q in range((-1) * charge_states, charge_states + 1):
                    parameters = {}
                    parameters['asr.relax'] = {'chargestate': q}
                    parameters['asr.gs@calculate'] = {'chargestate': q}
                    charge_string = 'charge_{}'.format(q)
                    charge_dict[charge_string] = {'structure': vacancy,
                                                  'parameters': parameters}
                    temp_dict[string] = charge_dict
            finished_list.append(eq_pos[i])

    # incorporate anti-site defects
    finished_list = []
    if intrinsic:
        defect_list = []
        for i in range(len(structure)):
            symbol = structure[i].symbol
            if symbol not in defect_list:
                defect_list.append(symbol)
        for i in range(len(structure)):
            if not eq_pos[i] in finished_list:
                for element in defect_list:
                    if not structure[i].symbol == element:
                        defect = pristine.copy()
                        defect[i].symbol = element
                        string = 'defects.{0}_{1}{2}{3}.{4}_at_{5}{6}'.format(
                                 formula, N_x, N_y, N_z, element,
                                 wyckoffs[i], i)
                        charge_dict = {}
                        for q in range(
                                (-1) * charge_states,
                                charge_states + 1):
                            parameters = {}
                            parameters['asr.relax'] = {'chargestate': q}
                            parameters['asr.gs@calculate'] = {'chargestate': q}
                            charge_string = 'charge_{}'.format(q)
                            charge_dict[charge_string] = {
                                'structure': defect, 'parameters': parameters}
                            temp_dict[string] = charge_dict
                finished_list.append(eq_pos[i])

    # put together structure dict
    structure_dict['defects'] = temp_dict

    print('INFO: setting up {0} different defect supercell systems in '
          'charge states -{1}, ..., +{1}, as well as the pristine supercell '
          'system.'.format(len(structure_dict['defects']), charge_states))

    return structure_dict


def create_folder_structure(structure, structure_dict, chargestates,
                            intrinsic, vacancies, sc, max_lattice, is_2D):
    """Create folder for all configurations.

    Creates a folder for every configuration of the defect supercell in
    the following way:

    - see example directory tree in 'main()'
    - these each contain two files: 'unrelaxed.json' (the defect
      supercell structure), 'params.json' (the non-general parameters
      of each system)
    - the content of those folders can then be used to do further
      processing (e.g. relax the defect structure)
    """
    from ase.io import write
    from asr.core import write_json

    # create a json file for general parameters that are equivalent for all
    # the different defect systems
    if sc == [0, 0, 0]:
        pristine, N_x, N_y, N_z = setup_supercell(
            structure, max_lattice, is_2D)
    else:
        N_x = sc[0]
        N_y = sc[1]
        N_z = sc[2]
    gen_params = {}
    gen_params['chargestates'] = chargestates
    gen_params['is_2D'] = is_2D
    gen_params['supercell'] = [N_x, N_y, N_z]
    gen_params['intrinsic'] = intrinsic
    gen_params['vacancies'] = vacancies
    write_json('general_parameters.json', gen_params)

    # then, create a seperate folder for each possible defect
    # configuration of this parent folder, as well as the pristine
    # supercell system
    for element in structure_dict:
        folder_name = element
        try:
            if not folder_name == 'defects':
                Path(folder_name).mkdir()
        except FileExistsError:
            print('WARNING: folder ("{0}") already exists in this '
                  f'directory. Skip creating it.'.format(folder_name))
        if structure_dict[element].get('structure') is not None:
            struc = structure_dict[element].get('structure')
            params = structure_dict[element].get('parameters')
            try:
                write(folder_name + '/unrelaxed.json', struc)
                write_json(folder_name + '/params.json', params)
            except FileExistsError:
                print('WARNING: files already exist inside this folder.')
        else:
            sub_dict = structure_dict[element]
            j = 0
            for sub_element in sub_dict:
                defect_name = [key for key in sub_dict.keys()]
                defect_folder_name = defect_name[j]
                j = j + 1
                try:
                    Path(defect_folder_name).mkdir()
                except FileExistsError:
                    print(
                        'WARNING: folder ("{0}") already exists in this '
                        f'directory. Skip creating '
                        f'it.'.format(defect_folder_name))
                for i in range((-1) * chargestates, chargestates + 1):
                    charge_name = 'charge_{}'.format(i)
                    charge_folder_name = defect_folder_name + '/' + charge_name
                    try:
                        Path(charge_folder_name).mkdir()
                    except FileExistsError:
                        print(
                            'WARNING: folder ("{0}") already exists in this '
                            f'directory. Skip creating '
                            f'it.'.format(charge_folder_name))
                    struc = sub_dict[sub_element].get(
                        charge_name).get('structure')
                    params = sub_dict[sub_element].get(
                        charge_name).get('parameters')
                    write(charge_folder_name + '/unrelaxed.json', struc)
                    write_json(charge_folder_name + '/params.json', params)

    return None


def collect_data():
    return None


# def webpanel(row, key_descriptions):
#    from asr.browser import fig, table
#
#    if 'something' not in row.data:
#        return None, []
#
#    table1 = table(row,
#                   'Property',
#                   ['something'],
#                   kd=key_descriptions)
#    panel = ('Title',
#             [[fig('something.png'), table1]])
#    things = [(create_plot, ['something.png'])]
#    return panel, things


# def create_plot(row, fname):
#    import matplotlib.pyplot as plt
#
#    data = row.data.something
#    fig = plt.figure()
#    ax = fig.gca()
#    ax.plot(data.things)
#    plt.savefig(fname)


if __name__ == '__main__':
    main.cli()



from asr.core import command, option, argument


@command('asr.database.clonetree')
@argument('patterns', nargs=-1, required=False, metavar='PATTERN')
@argument('destination', metavar='DESTDIR')
@argument('source', metavar='SRCDIR')
@option('--copy/--symlink', is_flag=True)
@option('--map-files')
@option('--dont-contain')
@option('--must-contain')
@option('--dry-run')
@option('--glob-pattern')
def main(source, destination, patterns,
         copy=False, map_files=None, dont_contain=None,
         must_contain=None, dry_run=False, glob_pattern='**/'):
    """Tool for copying or symlinking a tree of files."""
    import fnmatch
    from pathlib import Path
    from typing import List, Tuple
    from click import progressbar
    from os import makedirs

    print(f'Clone {source} to {destination}')
    if patterns:
        string = ', '.join(patterns)
        print(f'Patterns: {string}')

    if must_contain:
        must_contain = must_contain.split(',')
    else:
        must_contain = []

    if dont_contain:
        dont_contain = dont_contain.split(',')
    else:
        dont_contain = []

    source = Path(source)
    destination = Path(destination)

    if not patterns:
        patterns = ['*']

    log: List[Tuple[Path, Path]] = []
    mkdir: List[Path] = []
    errors = []

    def item_show_func(item):
        return str(item)

    with progressbar(source.glob(glob_pattern),
                     label='Searching for files and folders',
                     item_show_func=item_show_func) as bar:
        for srcdir in bar:
            destdir = destination / srcdir.relative_to(source)

            dirfiles = list(srcdir.glob('*'))
            dirfilenames = [srcfile.name for srcfile in dirfiles]

            # Directory has to contain all these files
            contains = [True if name in dirfilenames else False
                        for name in must_contain]
            if not all(contains):
                continue

            # Directory must not contain these files
            not_contains = [True if name in dirfilenames else False
                            for name in dont_contain]
            if any(not_contains):
                continue

            for srcfile in dirfiles:
                if srcfile.is_file():
                    destfile = destdir / srcfile.name
                    if any([fnmatch.fnmatch(srcfile.name, pattern)
                            for pattern in patterns]):
                        if destfile.is_file():
                            continue
                        log.append((srcfile, destfile))

            mkdir.append(destdir)

    if len(errors) > 0:
        for error in errors:
            print(error)
        raise AssertionError

    if not dry_run:
        with progressbar(mkdir,
                         label=f'Creating {len(mkdir)} folders') as bar:
            for destdir in bar:
                if destdir.is_dir():
                    continue
                makedirs(str(destdir))
    else:
        print(f'Would create {len(mkdir)} folders')

    if copy:
        if not dry_run:
            print(f'Copying {len(log)} files')
            with progressbar(log) as bar:
                for srcfile, destfile in bar:
                    destfile.write_bytes(srcfile.read_bytes())
        else:
            print(f'Would copy {len(copy)} files')
    else:
        if not dry_run:
            with progressbar(log, label=f'symlinking {len(log)} files') as bar:
                for srcfile, destfile in bar:
                    destfile.symlink_to(srcfile.resolve())
        else:
            print(f'Would create {len(mkdir)} folders and '
                  f'symlink {len(log)} files')

    # Finally we allow for some postprocessing or massaging of the
    # created folders if some files have to be swapped around
    from asr.core import chdir
    if map_files:
        mapping = [tmp.split('->') for tmp in map_files.split(',')]

        with progressbar(mkdir,
                         label=f'Remapping files') as bar:
            for destdir in bar:
                with chdir(destdir):
                    for orig, replace in mapping:
                        for src in Path('.').glob('*'):
                            if orig not in src.name:
                                continue
                            dest = Path(src.name.replace(orig, replace))
                            if Path(dest).is_file():
                                Path(dest).unlink()
                            dest.symlink_to(src.resolve())


if __name__ == '__main__':
    main.cli()



from asr.core import command, option


class NoGapError(Exception):
    pass


@command('asr.emasses',
         requires=['gs.gpw', 'results-asr.magnetic_anisotropy.json'],
         dependencies=['asr.structureinfo',
                       'asr.magnetic_anisotropy'],
         creates=['em_circle_vb_soc.gpw', 'em_circle_cb_soc.gpw'])
@option('--gpwfilename', type=str,
        help='GS Filename')
def refine(gpwfilename='gs.gpw'):
    """Take a bandstructure and calculate more kpts around the vbm and cbm."""
    from asr.utils.gpw2eigs import gpw2eigs
    from ase.dft.bandgap import bandgap
    from asr.magnetic_anisotropy import get_spin_axis
    import os.path
    socs = [True]

    for soc in socs:
        theta, phi = get_spin_axis()
        eigenvalues, efermi = gpw2eigs(gpw=gpwfilename, soc=soc,
                                       theta=theta, phi=phi)
        gap, _, _ = bandgap(eigenvalues=eigenvalues, efermi=efermi,
                            output=None)

        for bt in ['vb', 'cb']:
            name = get_name(soc=soc, bt=bt)
            gpw2 = name + '.gpw'

            if not gap > 0:
                raise NoGapError('Gap was zero')
            if os.path.exists(gpw2):
                continue
            gpwrefined = preliminary_refine(gpw=gpwfilename, soc=soc, bandtype=bt)
            nonsc_sphere(gpw=gpwrefined, soc=soc, bandtype=bt)


def get_name(soc, bt):
    return 'em_circle_{}_{}'.format(bt, ['nosoc', 'soc'][soc])


def preliminary_refine(gpw='gs.gpw', soc=True, bandtype=None):
    from gpaw import GPAW
    import numpy as np
    from asr.utils.gpw2eigs import gpw2eigs
    from ase.dft.bandgap import bandgap
    from asr.magnetic_anisotropy import get_spin_axis
    # Get calc and current kpts
    calc = GPAW(gpw, txt=None)
    ndim = calc.atoms.pbc.sum()
    k_kc = calc.get_ibz_k_points()
    cell_cv = calc.atoms.get_cell()

    # Find energies and VBM/CBM
    theta, phi = get_spin_axis()
    e_skn, efermi = gpw2eigs(gpw, soc=soc, theta=theta, phi=phi)
    if e_skn.ndim == 2:
        e_skn = e_skn[np.newaxis]
    _, (s1, k1, n1), (s2, k2, n2) = bandgap(eigenvalues=e_skn, efermi=efermi,
                                            output=None)
    # Make a sphere of kpts of high density
    nkpts = max(int(e_skn.shape[1]**(1 / ndim)), 19)
    nkpts = nkpts + (1 - (nkpts % 2))
    assert nkpts % 2 != 0
    ksphere = kptsinsphere(cell_cv, npoints=nkpts,
                           erange=500e-3, m=1.0,
                           dimensionality=ndim)

    # Position sphere around VBM if bandtype == vb
    # Else around CBM
    if bandtype == 'vb':
        newk_kc = k_kc[k1] + ksphere
    elif bandtype == 'cb':
        newk_kc = k_kc[k2] + ksphere
    else:
        raise ValueError(f'Bandtype {bandtype} not recognized')

    # Calculate energies for new k-grid
    fname = '_refined'
    calc.set(kpts=newk_kc,
             symmetry='off',
             txt=fname + '.txt')
    atoms = calc.get_atoms()
    atoms.get_potential_energy()
    calc.write(fname + '.gpw')
    return fname + '.gpw'


def nonsc_sphere(gpw='gs.gpw', soc=False, bandtype=None):
    """Non sc calculation for kpts in a sphere around the VBM/CBM.

    Writes the files:

    * em_circle_vb_soc.gpw
    * em_circle_cb_soc.gpw
    * em_circle_vb_nosoc.gpw
    * em_circle_cb_nosoc.gpw

    Parameters
    ----------
    gpw: str
        gpw filename
    soc: bool
        spinorbit coupling
    bandtype: None or 'cb' or 'vb'
        Which bandtype do we do calculations for, if None is done for
        for both cb and vb
    """
    from gpaw import GPAW
    import numpy as np
    from asr.utils.gpw2eigs import gpw2eigs
    from ase.dft.bandgap import bandgap
    from asr.magnetic_anisotropy import get_spin_axis
    calc = GPAW(gpw, txt=None)
    ndim = calc.atoms.pbc.sum()

    # Check that 1D: Only z-axis, 2D: Only x- and y-axis
    if ndim == 1:
        pbc = calc.atoms.pbc
        assert not pbc[0] and not pbc[1] and pbc[2]
    elif ndim == 2:
        pbc = calc.atoms.pbc
        assert pbc[0] and pbc[1] and not pbc[2]

    k_kc = calc.get_ibz_k_points()
    cell_cv = calc.atoms.get_cell()

    kcirc_kc = kptsinsphere(cell_cv, dimensionality=ndim)

    theta, phi = get_spin_axis()
    e_skn, efermi = gpw2eigs(gpw, soc=soc, theta=theta, phi=phi)
    if e_skn.ndim == 2:
        e_skn = e_skn[np.newaxis]

    _, (s1, k1, n1), (s2, k2, n2) = bandgap(eigenvalues=e_skn, efermi=efermi,
                                            output=None)

    k1_c = k_kc[k1]
    k2_c = k_kc[k2]

    bandtypes, ks = get_bt_ks(bandtype, k1_c, k2_c)

    for bt, k_c in zip(bandtypes, ks):
        name = get_name(soc=soc, bt=bt)
        calc.set(kpts=kcirc_kc + k_c,
                 symmetry='off',
                 txt=name + '.txt')
        atoms = calc.get_atoms()
        atoms.get_potential_energy()
        calc.write(name + '.gpw')


def kptsinsphere(cell_cv, npoints=9, erange=1e-3, m=1.0, dimensionality=3):
    import numpy as np
    from ase.units import Hartree, Bohr
    from ase.dft.kpoints import kpoint_convert

    a = np.linspace(-1, 1, npoints)
    X, Y, Z = np.meshgrid(a, a, a)

    na = np.logical_and
    if dimensionality == 2:
        indices = na(X**2 + Y**2 <= 1.0, Z == 0)
    elif dimensionality == 1:
        indices = na(Z**2 <= 1.0, na(X == 0, Y == 0))
    else:
        indices = X**2 + Y**2 + Z**2 <= 1.0

    x, y, z = X[indices], Y[indices], Z[indices]
    kpts_kv = np.vstack([x, y, z]).T
    kr = np.sqrt(2 * m * erange / Hartree)
    kpts_kv *= kr
    kpts_kv /= Bohr
    kpts_kc = kpoint_convert(cell_cv=cell_cv, ckpts_kv=kpts_kv)
    return kpts_kc

    # print(kpts_kv)

    # a = np.linspace(-1, 1, npoints)
    # X, Y, Z = np.meshgrid(a, a, a)
    # indices = X**2 + Y**2 + Z**2 <= 1.0
    # sh = X.shape
    # x, y, z = X[indices], Y[indices], Z[indices]
    # kpts_kv = np.vstack([x, y * yfactor, z * zfactor]).T


def get_bt_ks(bandtype, k1_c, k2_c):
    if bandtype is None:
        bandtypes = ('vb', 'cb')
        ks = (k1_c, k2_c)
    elif bandtype == 'vb':
        bandtypes = ('vb',)
        ks = (k1_c, )
    elif bandtype == 'cb':
        bandtypes = ('cb', )
        ks = (k2_c, )
    return bandtypes, ks


def get_ndims_from_row(row):
    import numpy as np
    results = row.data.get('results-asr.emasses.json')

    for k in results.keys():
        if '(' in k and ')' in k:
            count_dirs = 0
            for k2 in results[k].keys():
                if 'effmass' in k2:
                    value = results[k][k2]
                    if not (value is None) and not np.isnan(value):
                        count_dirs += 1
            break
    return count_dirs


def convert_key_to_tuple(key):
    k = key.replace("(", "").replace(")", "")
    ks = k.split(",")
    ks = [k.strip() for k in ks]
    ks = [int(k) for k in ks]
    return tuple(ks)


def get_emass_dict_from_row(row):
    import numpy as np
    results = row.data.get('results-asr.emasses.json')

    cb_indices = []
    vb_indices = []
    for k in results.keys():
        if '(' in k and ')' in k:
            for k2 in results[k].keys():
                if 'nosoc' in k2:
                    break

                if 'vb_soc_effmass' in k2:
                    vb_indices.append(k)
                    break
                elif 'cb_soc_effmass' in k2:
                    cb_indices.append(k)
                    break

    cb_indices = [(k, convert_key_to_tuple(k)[1]) for k in cb_indices]
    vb_indices = [(k, convert_key_to_tuple(k)[1]) for k in vb_indices]

    ordered_cb_indices = sorted(cb_indices, key=lambda el: el[1])
    ordered_vb_indices = sorted(vb_indices, key=lambda el: -el[1])

    def get_the_dict(ordered_indices, name, offset_sym):
        my_dict = {}
        for offset_num, (key, band_number) in enumerate(ordered_indices):
            data = results[key]
            direction = 0
            for k in data.keys():
                if 'effmass' in k:
                    mass = data[k]
                    if mass is not None and not np.isnan(mass):
                        direction += 1
                        mass_str = str(round(abs(mass) * 100) / 100)
                        if offset_num == 0:
                            my_dict[f'{name}, direction {direction}'] = \
                                f'{mass_str} m<sub>e</sub>'
                        else:
                            my_dict['{} {} {}, direction {}'.format(
                                name, offset_sym,
                                offset_num, direction)] = \
                                f'{mass_str} m<sub>e</sub>'
        return my_dict

    electron_dict = get_the_dict(ordered_cb_indices, 'CB', '+')
    hole_dict = get_the_dict(ordered_vb_indices, 'VB', '-')

    return electron_dict, hole_dict


def get_range(mass, _erange):
    from ase.units import Ha, Bohr
    return (2 * mass * _erange / Ha) ** 0.5 / Bohr


def make_the_plots(row, *args):
    from ase.dft.kpoints import kpoint_convert, labels_from_kpts
    from ase.units import Bohr, Ha
    import matplotlib.pyplot as plt
    import numpy as np
    from asr.database.browser import fig as asrfig

    results = row.data.get('results-asr.emasses.json')
    efermi = row.efermi
    sdir = row.get('spin_axis', 'z')
    cell_cv = row.cell

    reference = row.get('evac', efermi)
    label = r'E_\mathrm{vac}' if 'evac' in row else r'E_\mathrm{F}'
    columns = []
    cb_fnames = []
    vb_fnames = []

    vb_indices = []
    cb_indices = []

    for spin_band_str, data in results.items():
        if '__' in spin_band_str or not isinstance(data, dict):
            continue
        is_w_soc = check_soc(data)
        if not is_w_soc:
            continue
        for k in data.keys():
            if 'effmass' in k and 'vb' in k:
                vb_indices.append(spin_band_str)
                break
            if 'effmass' in k and 'cb' in k:
                cb_indices.append(spin_band_str)
                break

    cb_masses = {}
    vb_masses = {}

    for cb_key in cb_indices:
        data = results[cb_key]
        masses = []
        for k in data.keys():
            if 'effmass' in k:
                masses.append(data[k])
        tuple_key = convert_key_to_tuple(cb_key)
        cb_masses[tuple_key] = masses

    for vb_key in vb_indices:
        data = results[vb_key]
        masses = []
        for k in data.keys():
            if 'effmass' in k:
                masses.append(data[k])
        tuple_key = convert_key_to_tuple(vb_key)
        vb_masses[tuple_key] = masses

    erange = 0.05

    plt_count = 0
    for direction in range(3):
        y1 = None
        y2 = None
        my_range = None
        # CB plots
        fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(6.4, 2.8),
                                 sharey=True,
                                 gridspec_kw={'width_ratios': [1]})

        should_plot = True
        for cb_key in cb_indices[0:1]:
            cb_tuple = convert_key_to_tuple(cb_key)
            # Save something
            data = results[cb_key]
            fit_data_list = data['cb_soc_bzcuts']
            if direction >= len(fit_data_list):
                should_plot = False
                continue

            mass = cb_masses[cb_tuple][direction]
            fit_data = fit_data_list[direction]
            ks = fit_data['kpts_kc']
            bt = fit_data['bt']
            xk, _, _ = labels_from_kpts(kpts=ks, cell=cell_cv)
            xk -= xk[-1] / 2
            kpts_kv = kpoint_convert(cell_cv=cell_cv, skpts_kc=ks)
            kpts_kv *= Bohr

            e_k = fit_data['e_k'] - reference
            e_km = fit_data['e_km'] - reference
            sz_km = fit_data['spin_km']
            emodel_k = (xk * Bohr) ** 2 / (2 * mass) * Ha - reference
            emodel_k += np.min(e_k) - np.min(emodel_k)

            for im in range(e_km.shape[1]):
                things = axes.scatter(xk, e_km[:, im], c=sz_km[:, im], vmin=-1, vmax=1)
            axes.plot(xk, emodel_k, c='r', ls='--')

            if y1 is None or y2 is None or my_range is None:
                y1 = np.min(emodel_k) - erange * 0.25
                y2 = np.min(emodel_k) + erange * 0.75
                axes.set_ylim(y1, y2)

                my_range = get_range(abs(mass), erange)
                axes.set_xlim(-my_range, my_range)

                cbar = fig.colorbar(things, ax=axes)
                cbar.set_label(rf'$\langle S_{sdir} \rangle$')
                cbar.set_ticks([-1, -0.5, 0, 0.5, 1])
                cbar.update_ticks()
            plt.locator_params(axis='x', nbins=3)
            axes.set_ylabel(r'$E-{}$ [eV]'.format(label))
            axes.set_title(f'{bt.upper()}, direction {direction + 1}')
            axes.set_xlabel(r'$\Delta k$ [1/$\mathrm{\AA}$]')
            plt.tight_layout()
        if should_plot:
            fname = args[plt_count]
            plt.savefig(fname)
            plt.close()

        # axes.clear()
        # VB plots
        y1 = None
        y2 = None
        my_range = None
        fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(6.4, 2.8),
                                 sharey=True,
                                 gridspec_kw={'width_ratios': [1]})

        for vb_key in vb_indices[0:1]:
            # Save something
            vb_tuple = convert_key_to_tuple(vb_key)
            data = results[vb_key]
            fit_data_list = data['vb_soc_bzcuts']
            if direction >= len(fit_data_list):
                continue

            mass = vb_masses[vb_tuple][direction]
            fit_data = fit_data_list[direction]
            ks = fit_data['kpts_kc']
            bt = fit_data['bt']
            e_k = fit_data['e_k'] - reference
            e_km = fit_data['e_km'] - reference
            sz_km = fit_data['spin_km']
            xk2, y, y2 = labels_from_kpts(kpts=ks, cell=cell_cv, eps=1)
            xk2 -= xk2[-1] / 2

            kpts_kv = kpoint_convert(cell_cv=cell_cv, skpts_kc=ks)
            kpts_kv *= Bohr

            emodel_k = (xk2 * Bohr) ** 2 / (2 * mass) * Ha - reference
            emodel_k += np.max(e_k) - np.max(emodel_k)

            for im in range(e_km.shape[1]):
                things = axes.scatter(xk2, e_km[:, im], c=sz_km[:, im], vmin=-1, vmax=1)
            axes.plot(xk2, emodel_k, c='r', ls='--')

            if y1 is None or y2 is None or my_range is None:
                y1 = np.max(emodel_k) - erange * 0.75
                y2 = np.max(emodel_k) + erange * 0.25
                axes.set_ylim(y1, y2)

                my_range = get_range(abs(mass), erange)
                axes.set_xlim(-my_range, my_range)

                cbar = fig.colorbar(things, ax=axes)
                cbar.set_label(rf'$\langle S_{sdir} \rangle$')
                cbar.set_ticks([-1, -0.5, 0, 0.5, 1])
                cbar.update_ticks()
            plt.locator_params(axis='x', nbins=3)
            axes.set_ylabel(r'$E-{}$ [eV]'.format(label))
            axes.set_title(f'{bt.upper()}, direction {direction + 1}')
            axes.set_xlabel(r'$\Delta k$ [1/$\mathrm{\AA}$]')
            plt.tight_layout()
        if should_plot:
            nplts = len(args)
            fname = args[plt_count + nplts // 2]
            plt.savefig(fname)
            plt.close()

        plt_count += 1
    # Make final column with table of numerical vals

    # Make table

    # Algo:
    # Loop through directions, each direction is a column
    # For direction i, loop through cbs and plot on fig
    # -- Plot also quadratic fit from curvature/effective mass value
    # For direction i, loop through vbs and plot on fig
    # Make a final column containing a table with the numerical values
    # for the effective masses
    assert len(cb_fnames) == len(vb_fnames), \
        'Num cb plots: {}\nNum vb plots: {}'.format(
        len(cb_fnames), len(vb_fnames))

    num_cols = len(cb_fnames)

    for j in range(num_cols):
        cb_fname = cb_fnames[j]
        vb_fname = vb_fnames[j]
        col = [asrfig(cb_fname), asrfig(vb_fname)]
        columns.append(col)

    return


def custom_table(values_dict, title):
    table = {'type': 'table',
             'header': [title, 'Value']}

    rows = []
    for k in values_dict.keys():
        rows.append((k, values_dict[k]))

    table['rows'] = rows
    return table


def webpanel(row, key_descriptions):
    # from asr.database.browser import table

    columns, fnames = create_columns_fnames(row)

    electron_dict, hole_dict = get_emass_dict_from_row(row)

    electron_table = custom_table(electron_dict, 'Electron effective mass')
    hole_table = custom_table(hole_dict, 'Hole effective mass')
    columns[0].append(electron_table)
    columns[1].append(hole_table)

    panel = {'title': 'Effective masses (PBE)',
             'columns': columns,
             'plot_descriptions':
             [{'function': make_the_plots,
               'filenames': fnames
               }],
             'sort': 12}
    return [panel]


def create_columns_fnames(row):
    from asr.database.browser import fig as asrfig

    results = row.data.get('results-asr.emasses.json')

    cb_fnames = []
    vb_fnames = []

    vb_indices = []
    cb_indices = []

    for spin_band_str, data in results.items():
        if '__' in spin_band_str or not isinstance(data, dict):
            continue
        is_w_soc = check_soc(data)
        if not is_w_soc:
            continue
        for k in data.keys():
            if 'effmass' in k and 'vb' in k:
                vb_indices.append(spin_band_str)
                break
            if 'effmass' in k and 'cb' in k:
                cb_indices.append(spin_band_str)
                break

    cb_masses = {}
    vb_masses = {}

    for cb_key in cb_indices:
        data = results[cb_key]
        masses = []
        for k in data.keys():
            if 'effmass' in k:
                masses.append(data[k])
        tuple_key = convert_key_to_tuple(cb_key)
        cb_masses[tuple_key] = masses

    for vb_key in vb_indices:
        data = results[vb_key]
        masses = []
        for k in data.keys():
            if 'effmass' in k:
                masses.append(data[k])
        tuple_key = convert_key_to_tuple(vb_key)
        vb_masses[tuple_key] = masses

    for direction in range(3):
        should_plot = True
        for cb_key in cb_indices:
            data = results[cb_key]
            fit_data_list = data['cb_soc_bzcuts']
            if direction >= len(fit_data_list):
                should_plot = False
                continue
        if should_plot:
            fname = 'cb_dir_{}.png'.format(direction)
            cb_fnames.append(fname)

            fname = 'vb_dir_{}.png'.format(direction)
            vb_fnames.append(fname)

    assert len(cb_fnames) == len(vb_fnames), \
        'Num cb plots: {}\nNum vb plots: {}'.format(
        len(cb_fnames), len(vb_fnames))

    num_figs = len(cb_fnames)

    columns = [[], []]
    for j in range(num_figs):
        cb_fname = cb_fnames[j]
        vb_fname = vb_fnames[j]

        columns[0].append(asrfig(cb_fname))
        columns[1].append(asrfig(vb_fname))

    return columns, cb_fnames + vb_fnames


def check_soc(spin_band_dict):
    for k in spin_band_dict.keys():
        if 'effmass' in k and 'nosoc' in k:
            return False

    return True


@command('asr.emasses',
         requires=['em_circle_vb_soc.gpw', 'em_circle_cb_soc.gpw',
                   'gs.gpw', 'results-asr.structureinfo.json',
                   'results-asr.gs.json',
                   'results-asr.magnetic_anisotropy.json'],
         dependencies=['asr.emasses@refine',
                       'asr.gs@calculate',
                       'asr.gs',
                       'asr.structureinfo',
                       'asr.magnetic_anisotropy'],
         webpanel=webpanel)
@option('--gpwfilename', type=str,
        help='GS Filename')
def main(gpwfilename='gs.gpw'):
    from asr.utils.gpw2eigs import gpw2eigs
    from ase.dft.bandgap import bandgap
    from asr.magnetic_anisotropy import get_spin_axis
    import traceback
    socs = [True]

    good_results = {}
    for soc in socs:
        theta, phi = get_spin_axis()
        eigenvalues, efermi = gpw2eigs(gpw=gpwfilename, soc=soc,
                                       theta=theta, phi=phi)
        gap, _, _ = bandgap(eigenvalues=eigenvalues, efermi=efermi,
                            output=None)
        if not gap > 0:
            raise NoGapError('Gap was zero')
        for bt in ['vb', 'cb']:
            name = get_name(soc=soc, bt=bt)
            gpw2 = name + '.gpw'
            try:
                masses = embands(gpw2,
                                 soc=soc,
                                 bandtype=bt,
                                 efermi=efermi)

                # This function modifies the last argument
                unpack_masses(masses, soc, bt, good_results)
            except ValueError:
                tb = traceback.format_exc()
                print(gpw2 + ':\n' + '=' * len(gpw2) + '\n', tb)
            # else:
            #     _savemass(soc=soc, bt=bt, mass=masses)

    return good_results


def unpack_masses(masses, soc, bt, results_dict):
    # Structure of 'masses' object:
    # There is an 'indices' key which tells you at which spin-k indices
    # effective masses have been calculated
    # At a given index there is saved the dict returned by the em function:
    # out = dict(mass_u=masses, eigenvectors_vu=vecs,
    #            ke_v=kmax,
    #            c=c,
    #            r=r)
    # We want to flatten this structure so that results_dict contains
    # the masses directly
    import numpy as np

    for ind in masses['indices']:
        out_dict = masses[ind]
        index = str(ind)
        socpre = 'soc' if soc else 'nosoc'
        prefix = bt + '_' + socpre + '_'
        offset = out_dict['offset']

        results_dict[index] = {}

        mass_u = list(out_dict['mass_u'])

        for u, m in enumerate(mass_u):
            if np.isnan(m):
                mass_u[u] = None

        results_dict[index][prefix + 'effmass_dir1'] = mass_u[0]
        results_dict[index][prefix + 'effmass_dir2'] = mass_u[1]
        results_dict[index][prefix + 'effmass_dir3'] = mass_u[2]

        if offset == 0:
            results_dict[f'emass_{bt}_dir1'] = mass_u[0]
            results_dict[f'emass_{bt}_dir2'] = mass_u[1]
            results_dict[f'emass_{bt}_dir3'] = mass_u[2]

        vecs = out_dict['eigenvectors_vu']
        results_dict[index][prefix + 'eigenvectors_vdir1'] = vecs[:, 0]
        results_dict[index][prefix + 'eigenvectors_vdir2'] = vecs[:, 1]
        results_dict[index][prefix + 'eigenvectors_vdir3'] = vecs[:, 2]
        results_dict[index][prefix + 'spin'] = ind[0]
        results_dict[index][prefix + 'bandindex'] = ind[1]
        results_dict[index][prefix + 'kpt_v'] = out_dict['ke_v']
        results_dict[index][prefix + 'fitcoeff'] = out_dict['c']
        results_dict[index][prefix + '2ndOrderFit'] = out_dict['c2']
        results_dict[index][prefix + 'mass_u'] = mass_u
        results_dict[index][prefix + 'bzcuts'] = out_dict['bs_along_emasses']
        results_dict[index][prefix + 'fitkpts_kv'] = out_dict['fitkpts_kv']
        results_dict[index][prefix + 'fite_k'] = out_dict['fite_k']
        results_dict[index][prefix + '2ndOrderr2'] = out_dict['r2']
        results_dict[index][prefix + '3rdOrderr2'] = out_dict['r']
        results_dict[index][prefix + '2ndOrderMAE'] = out_dict['mae2']
        results_dict[index][prefix + '3rdOrderMAE'] = out_dict['mae3']
        results_dict[index][prefix + 'wideareaMAE'] = out_dict['wideareaMAE']


def embands(gpw, soc, bandtype, efermi=None, delta=0.1):
    """Effective masses for bands within delta of extrema.

    Parameters
    ----------
        gpw: str
            name of gpw filename
        soc: bool
            include spin-orbit coupling
        bandtype: 'vb' or 'cb'
            type of band
        efermi: float, optional
            fermi level (takes it from gpw if None)
        delta: float, optional
            bands within this value (in eV) is included in the em fit
            default is 0.1 eV

    """
    from gpaw import GPAW
    from asr.utils.gpw2eigs import gpw2eigs
    import numpy as np
    from ase.dft.kpoints import kpoint_convert
    from ase.units import Bohr, Hartree
    from asr.magnetic_anisotropy import get_spin_axis
    calc = GPAW(gpw, txt=None)
    ndim = calc.atoms.pbc.sum()

    theta, phi = get_spin_axis()
    e_skn, efermi2 = gpw2eigs(gpw, soc=soc, theta=theta, phi=phi)
    if efermi is None:
        efermi = efermi2
    if e_skn.ndim == 2:
        e_skn = e_skn[np.newaxis]
    vb_ind, cb_ind = get_vb_cb_indices(e_skn=e_skn, efermi=efermi, delta=delta)

    indices = vb_ind if bandtype == 'vb' else cb_ind
    atoms = calc.get_atoms()
    cell_cv = atoms.get_cell()
    ibz_kc = calc.get_ibz_k_points()

    ibz_kv = kpoint_convert(cell_cv=cell_cv, skpts_kc=ibz_kc)
    masses = {'indices': indices}
    for offset, b in enumerate(indices):
        e_k = e_skn[b[0], :, b[1]]
        masses[b] = em(kpts_kv=ibz_kv * Bohr,
                       eps_k=e_k / Hartree, bandtype=bandtype, ndim=ndim)

        calc_bs = calculate_bs_along_emass_vecs
        if offset == 0:
            nbands = len(indices)
        else:
            nbands = 1
        masses[b]['bs_along_emasses'] = calc_bs(masses[b],
                                                soc, bandtype, calc,
                                                spin=b[0],
                                                band=b[1],
                                                nbands=nbands)
        masses[b]['wideareaMAE'] = wideMAE(masses[b], bandtype,
                                           cell_cv)
        masses[b]['offset'] = offset

    return masses


def wideMAE(masses, bt, cell_cv, erange=25e-3):
    from ase.dft.kpoints import kpoint_convert
    from ase.units import Ha
    import numpy as np

    erange = erange / Ha

    maes = []
    for i, mass in enumerate(masses['mass_u']):
        if mass is np.nan or np.isnan(mass) or mass is None:
            continue

        fit_data = masses['bs_along_emasses'][i]
        c = masses['c']
        k_kc = fit_data['kpts_kc']
        k_kv = kpoint_convert(cell_cv=cell_cv, skpts_kc=k_kc)
        e_k = fit_data['e_k']

        if bt == "vb":
            ks = np.where(np.abs(e_k - np.max(e_k)) < erange)
            sk_kv = k_kv[ks]
        else:
            ks = np.where(np.abs(e_k - np.max(e_k)) < erange)
            sk_kv = k_kv[ks]

        emodel_k = evalmodel(sk_kv, c, thirdorder=True)

        mae = np.mean(np.abs(emodel_k - e_k[ks]))
        maes.append(mae)

    return maes


def calculate_bs_along_emass_vecs(masses_dict, soc,
                                  bt, calc,
                                  spin, band,
                                  erange=500e-3, npoints=91,
                                  nbands=1):
    from pathlib import Path
    from ase.units import Hartree, Bohr
    from ase.dft.kpoints import kpoint_convert
    from asr.utils.gpw2eigs import calc2eigs
    from asr.magnetic_anisotropy import get_spin_axis, get_spin_index
    from asr.core import file_barrier
    from gpaw import GPAW
    from gpaw.mpi import serial_comm
    import numpy as np
    cell_cv = calc.get_atoms().get_cell()

    results_dicts = []
    for u, mass in enumerate(masses_dict['mass_u']):
        if mass is np.nan or np.isnan(mass) or mass is None:
            continue
        identity = f'em_bs_spin={spin}_band={band}_bt={bt}_dir={u}_soc={soc}'
        name = f'{identity}.gpw'

        if not Path(name).is_file():
            with file_barrier([name]):
                # embzcut stuff
                kmax = np.sqrt(2 * abs(mass) * erange / Hartree)
                assert not np.isnan(kmax)
                kd_v = masses_dict['eigenvectors_vu'][:, u]
                assert not (np.isnan(kd_v)).any()
                k_kv = (np.linspace(-1, 1, npoints) * kmax * kd_v.reshape(3, 1)).T
                k_kv += masses_dict['ke_v']
                k_kv /= Bohr
                assert not (np.isnan(k_kv)).any()
                k_kc = kpoint_convert(cell_cv=cell_cv, ckpts_kv=k_kv)
                assert not (np.isnan(k_kc)).any()
                atoms = calc.get_atoms()
                for i, pb in enumerate(atoms.pbc):
                    if not pb:
                        k_kc[:, i] = 0
                assert not (np.isnan(k_kc)).any()
                calc.set(kpts=k_kc, symmetry='off',
                         txt=f'{identity}.txt', fixdensity=True)
                atoms.get_potential_energy()
                calc.write(name)

        calc_serial = GPAW(name, txt=None, communicator=serial_comm)
        k_kc = calc_serial.get_bz_k_points()
        theta, phi = get_spin_axis()
        e_km, _, s_kvm = calc2eigs(calc_serial, [0], soc=soc, return_spin=True,
                                   theta=theta, phi=phi)

        sz_km = s_kvm[:, get_spin_index(), :]

        dct = dict(bt=bt,
                   kpts_kc=k_kc,
                   e_k=e_km[:, band],
                   spin_k=sz_km[:, band])
        if nbands != 1:
            if bt == "vb":
                dct['e_km'] = e_km[:, band - nbands + 1:band + 1]
                dct['spin_km'] = sz_km[:, band - nbands + 1:band + 1]
            else:
                dct['e_km'] = e_km[:, band:band + nbands]
                dct['spin_km'] = sz_km[:, band:band + nbands]

        results_dicts.append(dct)

    return results_dicts


def get_vb_cb_indices(e_skn, efermi, delta):
    """Find CB and VB within a distance of delta of the CB and VB extrema.

    Parameters
    ----------
        e_skn: (ns, nk, nb)-shape ndarray
            eigenvalues
        efermi: float
            fermi level
        delta: float
            bands within delta of the extrema are included
    Returns
    -------
        vb_indices, cb_indices: [(spin, band), ..], [(spin, band), ...]
            spin and band indices (aka as SBandex) for VB and CB, respectively

    """
    import numpy as np
    from ase.dft.bandgap import bandgap
    if e_skn.ndim == 2:
        e_skn = e_skn[np.newaxis]
    gap, (s1, k1, n1), (s2, k2, n2) = bandgap(eigenvalues=e_skn,
                                              efermi=efermi, output=None)

    if not gap > 0:
        raise ValueError('Band gap is zero')

    cbm = e_skn[s2, k2, n2]
    vbm = e_skn[s1, k1, n1]

    cb_sn = e_skn[:, k2, n2:]
    vb_sn = e_skn[:, k1, :n1 + 1]
    cbs, cbn = np.where(cb_sn <= cbm + delta)
    cbn += n2
    cb_indices = list(zip(cbs, cbn))

    vbs, vbn = np.where(vb_sn >= vbm - delta)
    vb_indices = list(reversed(list(zip(vbs, vbn))))
    return vb_indices, cb_indices


def em(kpts_kv, eps_k, bandtype=None, ndim=3):
    """TODO: Do me.

    Parameters
    ----------
        kpts_kv: (nk, 3)-shape ndarray
            k-points in cartesian coordinates (in units of 1 / Bohr)
        eps_k: (nk,)-shape ndarray
            eigenvalues (in units of Hartree)
    Returns
    -------
        out: dct
            - effective masses in units of m_e
            - eigenvectors in cartesian coordinates
            - k-pot extremum in cartesian coordinates (units of 1 / Bohr)

    """
    import numpy as np
    c, r, rank, s, = fit(kpts_kv, eps_k, thirdorder=False)
    dxx = 2 * c[0]
    dyy = 2 * c[1]
    dzz = 2 * c[2]
    dxy = c[3]
    dxz = c[4]
    dyz = c[5]

    mae2 = np.mean(np.abs(eps_k - evalmodel(kpts_kv, c, thirdorder=False)))

    xm, ym, zm = get_2nd_order_extremum(c, ndim=ndim)
    ke2_v = np.array([xm, ym, zm])

    c3, r3, rank3, s3 = fit(kpts_kv, eps_k, thirdorder=True)

    mae3 = np.mean(np.abs(eps_k - evalmodel(kpts_kv, c3, thirdorder=True)))

    f3xx, f3yy, f3zz, f3xy = c3[:4]
    f3xz, f3yz, f3x, f3y = c3[4:8]
    f3z, f30, f3xxx, f3yyy = c3[8:12]
    f3zzz, f3xxy, f3xxz, f3yyx, f3yyz, f3zzx, f3zzy, f3xyz = c3[12:]

    if ndim == 2:
        def check_zero(v, i):
            assert np.allclose(v, 0), "Value was {} for index {}".format(v, i)

        zeros = [f3zz, f3xz, f3yz, f3z, f3zzz, f3xxz,
                 f3yyz, f3zzx, f3zzy, f3xyz]
        for i, z in enumerate(zeros):
            check_zero(z, i)

    extremum_type = get_extremum_type(dxx, dyy, dzz, dxy, dxz, dyz, ndim=ndim)
    # if bandtype == 'vb':
    #     assert extremum_type == 'max'
    # elif bandtype == 'cb':
    #     assert extremum_type == 'min'
    # else:
    #     raise NotImplementedError("Incorrect bandtype: {}".format(bandtype))
    xm, ym, zm = get_3rd_order_extremum(xm, ym, zm, c3,
                                        extremum_type, ndim=ndim)
    ke_v = np.array([xm, ym, zm])

    assert not (np.isnan(ke_v)).any()
    if ndim == 2:
        assert np.allclose(zm, 0), zm

    d3xx = (2 * f3xx) + (6 * f3xxx * xm) + (2 * f3xxy * ym) + (2 * f3xxz * zm)
    d3yy = (2 * f3yy) + (6 * f3yyy * ym) + (2 * f3yyx * xm) + (2 * f3yyz * zm)
    d3zz = (2 * f3zz) + (6 * f3zzz * zm) + (2 * f3zzx * xm) + (2 * f3zzy * ym)
    d3xy = f3xy + (2 * f3xxy * xm) + (2 * f3yyx * ym) + (f3xyz * zm)
    d3xz = f3xz + (2 * f3xxz * xm) + (2 * f3zzx * zm) + (f3xyz * ym)
    d3yz = f3yz + (2 * f3yyz * ym) + (2 * f3zzy * zm) + (f3xyz * xm)

    hessian3 = np.array([[d3xx, d3xy, d3xz],
                         [d3xy, d3yy, d3yz],
                         [d3xz, d3yz, d3zz]])

    v3_n, w3_vn = np.linalg.eigh(hessian3)
    assert not (np.isnan(w3_vn)).any()

    hessian = np.array([[dxx, dxy, dxz],
                        [dxy, dyy, dyz],
                        [dxz, dyz, dzz]])
    v2_n, vecs = np.linalg.eigh(hessian)

    mass_u = 1 / v3_n
    for u, v3 in enumerate(v3_n):
        if np.allclose(v3, 0):
            mass_u[u] = np.nan

    sort_args = np.argsort(mass_u)

    mass_u = mass_u[sort_args]
    w3_vn = w3_vn[:, sort_args]

    out = dict(mass_u=mass_u,
               eigenvectors_vu=w3_vn,
               ke_v=ke_v,
               c=c3,
               r=r3,
               mass2_u=1 / v2_n,
               eigenvectors2_vu=vecs,
               ke2_v=ke2_v,
               c2=c,
               r2=r,
               fitkpts_kv=kpts_kv,
               fite_k=eps_k,
               mae2=mae2,
               mae3=mae3)

    return out


def get_extremum_type(dxx, dyy, dzz, dxy, dxz, dyz, ndim=3):
    # Input: 2nd order derivatives at the extremum point
    import numpy as np
    if ndim == 3:
        hessian = np.array([[dxx, dxy, dxz],
                            [dxy, dyy, dyz],
                            [dxz, dyz, dzz]])
        vals, vecs = np.linalg.eigh(hessian)
        saddlepoint = not (np.sign(vals[0]) == np.sign(vals[1])
                           and np.sign(vals[0]) == np.sign(vals[2]))

        if saddlepoint:
            etype = 'saddlepoint'
        elif (vals < 0).all():
            etype = 'max'
        elif (vals > 0).all():
            etype = 'min'
        else:
            msg = 'Extremum type could not be' \
                  + 'found for hessian: {}'.format(hessian)
            raise ValueError(msg)
        return etype
    elif ndim == 2:
        # Assume x and y axis are the periodic directions
        hessian = np.array([[dxx, dxy],
                            [dxy, dyy]])
        det = np.linalg.det(hessian)
        if det < 0:
            etype = 'saddlepoint'
        elif dxx < 0 and dyy < 0:
            etype = 'max'
        elif dxx > 0 and dyy > 0:
            etype = 'min'
        else:
            raise ValueError('Extremum type could not'
                             + 'be determined for hessian: {}'.format(hessian))
        return etype
    elif ndim == 1:
        # Assume z axis is the periodic direction
        if dzz < 0:
            etype = 'max'
        else:
            etype = 'min'
        return etype


def get_2nd_order_extremum(c, ndim=3):
    import numpy as np
    # fit is
    # fxx x^2 + fyy y^2 + fzz z^2 +
    # fxy xy + fxz xz + fyz yz + fx x + fy y + fz z + f0
    assert len(c) == 10
    fxx, fyy, fzz, fxy, fxz, fyz, fx, fy, fz, f0 = c

    if ndim == 3:
        ma = np.array([[2 * fxx, fxy, fxz],
                       [fxy, 2 * fyy, fyz],
                       [fxz, fyz, 2 * fzz]])

        v = np.array([-fx, -fy, -fz])

        min_pos = np.linalg.solve(ma, v)

        return min_pos

    elif ndim == 2:
        # Assume x and y are periodic directions
        ma = np.array([[2 * fxx, fxy],
                       [fxy, 2 * fyy]])
        v = np.array([-fx, -fy])
        min_pos = np.linalg.solve(ma, v)

        return np.array([min_pos[0], min_pos[1], 0.0])

    elif ndim == 1:
        # Assume z is periodic direction
        return np.array([0.0, 0.0, -fz / (2 * fzz)])


def get_3rd_order_extremum(xm, ym, zm, c, extremum_type, ndim=3):
    if extremum_type == 'saddlepoint':
        return xm, ym, zm

    import numpy as np
    from scipy import optimize

    assert len(c) == 20

    if ndim == 3:
        def get_v(kpts):
            k = np.asarray(kpts)
            if k.ndim == 1:
                k = k[np.newaxis]
            return model(k)
    elif ndim == 2:
        # Assume x and y are periodic
        # Remove z-dependence
        def get_v(kpts):
            k = np.asarray(kpts)
            if k.ndim == 1:
                k = k[np.newaxis]
            m = model(k)
            return m
    elif ndim == 1:
        # Assume z is periodic
        # Remove x, y - dependence
        def get_v(kpts):
            k = np.asarray(kpts)
            if k.ndim == 1:
                k = k[np.newaxis]
            m = model(k)
            return m

    # We want to use a minimization function from scipy
    # so if the extremum type is a 'max' we need to multiply
    # the function by - 1
    if extremum_type == 'max':
        def func(v):
            return -1 * np.dot(get_v(v), c)
    else:
        def func(v):
            return np.dot(get_v(v), c)

    x0 = np.array([xm, ym, zm])
    x, y, z = optimize.fmin(func, x0=x0,
                            xtol=1.0e-15, ftol=1.0e-15, disp=False)

    model_deltaE = np.abs(func(np.array([x, y, z])) - func(x0))

    if model_deltaE > 1e-3:
        x = xm
        y = ym
        z = zm

    if ndim == 2:
        return x, y, 0
    elif ndim == 1:
        return 0, 0, z
    else:
        return x, y, z


def fit(kpts_kv, eps_k, thirdorder=False):
    import numpy.linalg as la
    A_kp = model(kpts_kv)
    if not thirdorder:
        A_kp = A_kp[:, :10]
    return la.lstsq(A_kp, eps_k, rcond=-1)


def model(kpts_kv):
    """Calculate simple third order model.

    Parameters
    ----------
        kpts_kv: (nk, 3)-shape ndarray
            units of (1 / Bohr)

    """
    import numpy as np
    k_kx, k_ky, k_kz = kpts_kv[:, 0], kpts_kv[:, 1], kpts_kv[:, 2]

    ones = np.ones(len(k_kx))

    A_dp = np.array([k_kx**2,
                     k_ky**2,
                     k_kz**2,
                     k_kx * k_ky,
                     k_kx * k_kz,
                     k_ky * k_kz,
                     k_kx,
                     k_ky,
                     k_kz,
                     ones,
                     k_kx**3,
                     k_ky**3,
                     k_kz**3,
                     k_kx**2 * k_ky,
                     k_kx**2 * k_kz,
                     k_ky**2 * k_kx,
                     k_ky**2 * k_kz,
                     k_kz**2 * k_kx,
                     k_kz**2 * k_ky,
                     k_kx * k_ky * k_kz]).T

    return A_dp


def evalmodel(kpts_kv, c_p, thirdorder=True):
    import numpy as np
    kpts_kv = np.asarray(kpts_kv)
    if kpts_kv.ndim == 1:
        kpts_kv = kpts_kv[np.newaxis]
    A_kp = model(kpts_kv)
    if not thirdorder:
        A_kp = A_kp[:, :10]
    return np.dot(A_kp, c_p)


if __name__ == '__main__':
    main.cli()



"""Template recipe."""
import json
from pathlib import Path
from asr.core import command, option


@command('asr.something')
@option('--number', type=int)
def main(number=5):
    """Calculate something."""
    something = calculate_something(number)
    results = {'number': number,
               'something': something}
    Path('something.json').write_text(json.dumps(results))
    return results


def calculate_something(number):
    return number + 2


def collect_data(atoms):
    path = Path('something.json')
    if not path.is_file():
        return {}, {}, {}
    # Read data:
    dct = json.loads(path.read_text())
    # Define key-value pairs, key descriptions and data:
    kvp = {'something': dct['something']}
    kd = {'something': ('Something', 'Longer description', 'unit')}
    data = {'something':
            {'stuff': 'more complicated data structures',
             'things': [0, 1, 2, 1, 0]}}
    return kvp, kd, data


def webpanel(row, key_descriptions):
    from asr.browser import fig, table

    if 'something' not in row.data:
        return None, []

    table1 = table(row,
                   'Property',
                   ['something'],
                   kd=key_descriptions)
    panel = ('Title',
             [[fig('something.png'), table1]])
    things = [(create_plot, ['something.png'])]
    return panel, things


def create_plot(row, fname):
    import matplotlib.pyplot as plt

    data = row.data.something
    fig = plt.figure()
    ax = fig.gca()
    ax.plot(data.things)
    plt.savefig(fname)


group = 'property'
creates = ['something.json']  # what files are created
dependencies = []  # no dependencies
resources = '1:10m'  # 1 core for 10 minutes
diskspace = 0  # how much diskspace is used
restart = 0  # how many times to restart

if __name__ == '__main__':
    main()



def restrict_spin_projection_2d(kpt, op_scc, s_vm):
    import numpy as np
    mirror_count = 0
    s_vm = s_vm.copy()
    for symm in op_scc:
        # Inversion symmetry forces spin degeneracy and 180 degree rotation
        # forces the spins to lie in plane
        if np.allclose(symm, -1 * np.eye(3)):
            s_vm[:] = 0
            continue
        vals, vecs = np.linalg.eigh(symm)
        # A mirror plane
        if np.allclose(np.abs(vals), 1) and np.allclose(np.prod(vals), -1):
            # Mapping k -> k, modulo a lattice vector
            if np.allclose(kpt % 1, (np.dot(symm, kpt)) % 1):
                mirror_count += 1
    # If we have two or more mirror planes,
    # then we must have a spin-degenerate
    # subspace
    if mirror_count >= 2:
        s_vm[2, :] = 0

    return s_vm



from asr.core import command


def get_reduced_formula(formula, stoichiometry=False):
    """Get reduced formula from formula.

    Returns the reduced formula corresponding to a chemical formula,
    in the same order as the original formula
    E.g. Cu2S4 -> CuS2

    Parameters
    ----------
    formula : str
    stoichiometry : bool
        If True, return the stoichiometry ignoring the
        elements appearing in the formula, so for example "AB2" rather than
        "MoS2"

    Returns
    -------
        A string containing the reduced formula.
    """
    from functools import reduce
    from math import gcd
    import string
    import re
    split = re.findall('[A-Z][^A-Z]*', formula)
    matches = [re.match('([^0-9]*)([0-9]+)', x)
               for x in split]
    numbers = [int(x.group(2)) if x else 1 for x in matches]
    symbols = [matches[i].group(1) if matches[i] else split[i]
               for i in range(len(matches))]
    divisor = reduce(gcd, numbers)
    result = ''
    numbers = [x // divisor for x in numbers]
    numbers = [str(x) if x != 1 else '' for x in numbers]
    if stoichiometry:
        numbers = sorted(numbers)
        symbols = string.ascii_uppercase
    for symbol, number in zip(symbols, numbers):
        result += symbol + number
    return result


def has_inversion(atoms, use_spglib=True):
    """Determine if atoms has inversion symmetry.

    Parameters
    ----------
    atoms: Atoms object
           atoms
    use_spglib: bool
           use spglib

    Returns
    -------
        out: bool
    """
    import numpy as np
    import spglib

    atoms2 = atoms.copy()
    atoms2.pbc[:] = True
    atoms2.center(axis=2)
    cell = (atoms2.cell.array,
            atoms2.get_scaled_positions(),
            atoms2.numbers)
    R = -np.identity(3, dtype=int)
    r_n = spglib.get_symmetry(cell, symprec=1.0e-3)['rotations']
    return np.any([np.all(r == R) for r in r_n])


def webpanel(row, key_descriptions):
    from asr.database.browser import table

    basictable = table(row, 'Structure info', [
        'crystal_prototype', 'class', 'spacegroup', 'spgnum', 'ICSD_id',
        'COD_id'
    ], key_descriptions, 2)
    basictable['columnwidth'] = 4
    rows = basictable['rows']
    codid = row.get('COD_id')
    if codid:
        # Monkey patch to make a link
        for tmprow in rows:
            href = ('<a href="http://www.crystallography.net/cod/'
                    + '{id}.html">{id}</a>'.format(id=codid))
            if 'COD' in tmprow[0]:
                tmprow[1] = href

    doi = row.get('doi')
    if doi:
        rows.append([
            'Monolayer reported DOI',
            '<a href="https://doi.org/{doi}" target="_blank">{doi}'
            '</a>'.format(doi=doi)
        ])

    row = ['Magnetic state', row.magstate]
    eltable = {'type': 'table',
               'header': ['Electronic properties', ''],
               'rows': [row],
               'columnwidth': 4}

    panel = {'title': 'Summary',
             'columns': [[basictable,
                          {'type': 'table', 'header': ['Stability', ''],
                           'rows': [],
                           'columnwidth': 4},
                          eltable],
                         [{'type': 'atoms'}, {'type': 'cell'}]],
             'sort': -1}
    return [panel]


tests = [{'description': 'Test SI.',
          'cli': ['asr run "setup.materials -s Si2"',
                  'ase convert materials.json structure.json',
                  'asr run "setup.params asr.gs@calculate:ecut 300 '
                  'asr.gs@calculate:kptdensity 2"',
                  'asr run structureinfo',
                  'asr run database.fromtree',
                  'asr run "database.browser --only-figures"']}]


@command('asr.structureinfo',
         tests=tests,
         requires=['structure.json'],
         webpanel=webpanel)
def main():
    """Get structural information of atomic structure.

    This recipe produces information such as the space group and magnetic
    state properties that requires only an atomic structure. This recipes read
    the atomic structure in `structure.json`.
    """
    import numpy as np
    from ase.io import read
    from pathlib import Path

    atoms = read('structure.json')
    info = {}

    folder = Path().cwd()
    info['folder'] = str(folder)

    # Determine magnetic state
    def get_magstate(a):
        magmom = a.get_magnetic_moment()
        if abs(magmom) > 0.02:
            return 'fm'

        magmoms = a.get_magnetic_moments()
        if abs(magmom) < 0.02 and abs(magmoms).max() > 0.1:
            return 'afm'

        # Material is essentially non-magnetic
        return 'nm'

    try:
        magstate = get_magstate(atoms)
    except RuntimeError:
        magstate = 'nm'
    info['magstate'] = magstate.upper()

    formula = atoms.get_chemical_formula(mode='metal')
    stoichimetry = get_reduced_formula(formula, stoichiometry=True)
    info['formula'] = formula
    info['stoichiometry'] = stoichimetry
    info['has_inversion_symmetry'] = has_inversion(atoms)

    # Calculate crystal prototype
    import spglib
    formula = atoms.symbols.formula
    cell = (atoms.cell.array,
            atoms.get_scaled_positions(),
            atoms.numbers)
    stoi = atoms.symbols.formula.stoichiometry()[0]
    dataset = spglib.get_symmetry_dataset(cell, symprec=1e-3,
                                          angle_tolerance=0.1)
    info['spglib_dataset'] = dataset
    sg = dataset['international']
    number = dataset['number']
    w = ''.join(sorted(set(dataset['wyckoffs'])))
    crystal_prototype = f'{stoi}-{number}-{w}'
    info['crystal_prototype'] = crystal_prototype
    info['spacegroup'] = sg
    info['spgnum'] = number

    # Set temporary uid.
    # Will be changed later once we know the prototype.
    info['is_magnetic'] = info['magstate'] != 'NM'

    if (atoms.pbc == [True, True, False]).all():
        info['cell_area'] = abs(np.linalg.det(atoms.cell[:2, :2]))

    info['__key_descriptions__'] = {
        'magstate': 'KVP: Magnetic state',
        'is_magnetic': 'KVP: Material is magnetic (Magnetic)',
        'cell_area': 'KVP: Area of unit-cell [Ang^2]',
        'has_invsymm': 'KVP: Inversion symmetry',
        'stoichiometry': 'KVP: Stoichiometry',
        'spacegroup': 'KVP: Space group',
        'spgnum': 'KVP: Space group number',
        'crystal_prototype': 'KVP: Crystal prototype'}

    return info


if __name__ == '__main__':
    main.cli()



import pytest

from .conftest import test_materials, get_webcontent


@pytest.mark.ci
@pytest.mark.parametrize('atoms', test_materials)
def test_hse(separate_folder, atoms, mockgpaw, mocker):
    import numpy as np
    from pathlib import Path
    atoms.write('structure.json')

    def non_self_consistent_eigenvalues(calc,
                                        xcname,
                                        n1,
                                        n2,
                                        kpt_indices=None,
                                        snapshot=None,
                                        ftol=42.0):
        Path(snapshot).write_text('{}')
        e_skn = calc.eigenvalues[np.newaxis, :, n1:n2]
        return e_skn, e_skn, e_skn

    mocker.patch('gpaw.hybrids.eigenvalues.non_self_consistent_eigenvalues',
                 create=True, new=non_self_consistent_eigenvalues)

    def vxc(calc, xc):
        return calc.eigenvalues[np.newaxis]

    mocker.patch('gpaw.xc.tools.vxc', create=True, new=vxc)
    from asr.hse import main
    main()
    get_webcontent('database.db')



from .conftest import test_materials
import pytest
import numpy as np


@pytest.mark.ci
@pytest.mark.parametrize("atoms", test_materials)
def test_setup_displacements_displace_atom(atoms):
    from asr.setup.displacements import displace_atom
    delta = 0.02
    pos_av = atoms.get_positions()
    for ia in range(len(atoms)):
        for iv in [0, 1, 2]:
            for sign in [-1, 1]:
                new_atoms = displace_atom(atoms, ia, iv, sign, delta)
                new_pos_av = new_atoms.get_positions()

                posav = pos_av[ia, iv]
                new_posav = new_pos_av[ia, iv]
                assert np.abs(posav - new_posav) == pytest.approx(delta)


@pytest.mark.ci
@pytest.mark.parametrize("atoms", test_materials)
def test_setup_displacements_get_all_displacements(atoms):
    from asr.setup.displacements import get_all_displacements
    displacements = list(get_all_displacements(atoms))

    test_disps = [(ia, iv, sign) for ia in range(len(atoms)) for iv in
                  range(3) for sign in [-1, 1]]

    for disp in test_disps:
        assert disp in displacements

    assert len(displacements) == len(test_disps)


@pytest.mark.ci
@pytest.mark.parametrize("atoms", test_materials)
def test_setup_displacements(separate_folder, atoms):
    from asr.setup.displacements import main as displacements

    atoms.write('structure.json')
    results = displacements()
    generated_folders = results['folders']
    assert 'displacements/0.01-0-+x' in generated_folders



import numpy as np


def raw_orbital_LDOS(calc, a, spin, angular, nbands=None):
    eigenvalues = calc.eigenvalues
    weights_xi = np.zeros(list(eigenvalues.shape) + [len(angular)]) + 1 / len(angular)
    return eigenvalues[:nbands], weights_xi



import pytest
from pytest import approx
from .conftest import test_materials
import numpy as np


@pytest.mark.ci
@pytest.mark.parametrize("atoms", test_materials)
def test_borncharges(separate_folder, mockgpaw, mocker, atoms):
    from gpaw import GPAW
    from asr.borncharges import main

    natoms = len(atoms)

    # Number of electrons on each atom
    Z_a = np.array([-2 if ia % 2 else -2 for ia in range(natoms)])
    positive_charge = 1

    # This controls the positive charge of the ions
    def _get_setup_nvalence(self, element_number):
        return positive_charge

    # The dipole moment is used for non-periodic directions
    def _get_dipole_moment(self):
        return np.dot(Z_a + positive_charge, self.atoms.get_positions())

    # This controls the electronic contribution to the berry phase
    def _get_berry_phases(self, dir=0, spin=0):
        phase_c = 2 * np.pi * np.dot(Z_a, self.spos_ac) / 2  # / 2 is for spin
        return [phase_c[dir]]

    mocker.patch.object(GPAW, '_get_setup_nvalence', new=_get_setup_nvalence)
    mocker.patch.object(GPAW, '_get_dipole_moment', new=_get_dipole_moment)
    mocker.patch.object(GPAW, '_get_berry_phases', new=_get_berry_phases)

    atoms.write('structure.json')
    results = main()

    Z_analytical_avv = np.array([(Z + positive_charge) * np.eye(3) for Z in Z_a])
    Z_avv = np.array(results['Z_avv'])
    assert Z_analytical_avv == approx(Z_avv)


@pytest.mark.acceptance_test
def test_gpaw_berry_get_berry_phases_integration(separate_folder):
    from .conftest import BN
    from asr.borncharges import main
    from asr.setup.params import main as setupparams

    calculator_params = {
        'name': 'gpaw',
        'mode': {'name': 'pw', 'ecut': 300},
        'xc': 'PBE',
        'basis': 'dzp',
        'kpts': {'density': 2.0, 'gamma': True},
        'occupations': {'name': 'fermi-dirac',
                        'width': 0.05},
        'convergence': {'bands': 'CBM+3.0'},
        'nbands': '100%',
        'txt': 'gs.txt',
        'charge': 0,
    }

    formal_calculator = {
        'name': 'gpaw',
        'mode': {'name': 'pw', 'ecut': 300},
        'xc': 'PBE',
        'basis': 'dzp',
        'kpts': {'density': 2.0},
        'occupations': {'name': 'fermi-dirac',
                        'width': 0.05},
        'txt': 'formalpol.txt',
        'charge': 0
    }
    setupparams(params={'asr.gs@calculate': {'calculator': calculator_params},
                        'asr.formalpolarization': {'calculator': formal_calculator}})
    BN.write('structure.json')

    results = main()

    ZB_vv = np.eye(3)
    diag = np.eye(3, dtype=bool)
    ZB_vv[diag] = [2.71, 2.71, 0.27]
    for Z_vv, sym in zip(results['Z_avv'], results['sym_a']):
        if sym == 'B':
            sign = 1
        else:
            sign = -1

        assert np.all(sign * Z_vv[diag] > 0)
        assert Z_vv[~diag] == pytest.approx(0, abs=0.001)
        assert Z_vv == pytest.approx(sign * ZB_vv, abs=0.5)



import numpy as np
from types import SimpleNamespace


class DielectricFunction:

    chi0 = SimpleNamespace(plasmafreq_vv=np.zeros((3, 3), float))

    def __init__(self, *args, **kwargs):
        pass

    def get_frequencies(self):
        return np.linspace(0, 10, 100)

    def get_polarizability(self, *args, **kwargs):
        alpha_w = np.zeros((100, ), float) + 1j * 0
        return alpha_w, alpha_w



from .conftest import test_materials
import pytest


@pytest.mark.ci
@pytest.mark.parametrize("atoms", test_materials)
def test_bandstructure_main(separate_folder, mockgpaw, atoms):
    from ase.io import write
    from asr.bandstructure import main
    write('structure.json', atoms)
    main()



import numpy as np


def get_spinorbit_eigenvalues(
        calc,
        bands=None,
        gw_kn=None,
        return_spin=False,
        return_wfs=False,
        scale=1.0,
        theta=0.0,
        phi=0.0,
):

    nk = len(calc.get_ibz_k_points())
    nspins = 2
    nbands = calc.get_number_of_bands()
    bands = list(range(nbands))

    e_ksn = np.array(
        [
            [
                calc.get_eigenvalues(kpt=k, spin=s)[bands]
                for s in range(nspins)
            ]
            for k in range(nk)
        ]
    )

    s_kvm = np.zeros((nk, 3, nbands * 2), float)
    s_kvm[:, 2, ::2] = 1
    s_kvm[:, 2, ::2] = -1
    e_km = e_ksn.reshape((nk, -1))
    if return_spin:
        return e_km.T, s_kvm
    else:
        return e_km.T


def get_anisotropy(*args, **kwargs):
    return 0



from asr.core import command, option


@command('asr.setup.materials',
         creates=['materials.json'])
@option('-s', '--selection', type=str,
        help='ASE DB selection string')
def main(selection=''):
    """Create database with materials from the ASR materials database.

    The ASR materials database currently contains all elementary and
    binary materials from the Nomad benchmarking database.

    The created materials will be saved into the database
    "materials.json".

    Examples
    --------
    Get all materials from database
    >>> asr run setup.materials
    In folder: . (1/1)
    Running asr.setup.materials(selection='')
    PWD /scratch/mogje/asr
    Created materials.json database containing 136 materials

    """
    from ase.db import connect
    from pathlib import Path

    dbname = str(Path(__file__).parent / 'testsystems.json')
    db = connect(dbname)
    rows = list(db.select(selection))

    nmat = len(rows)
    assert not Path('materials.json').is_file(), \
        'Database materials.json already exists!'

    print('PWD', Path().resolve())
    newdb = connect('materials.json')
    for row in rows:
        atoms = row.toatoms()
        kvp = row.key_value_pairs
        data = row.data
        newdb.write(atoms, key_value_pairs=kvp, data=data)
    print(f'Created materials.json database containing {nmat} materials')


group = 'setup'


if __name__ == '__main__':
    main.cli()



import pytest


@pytest.mark.ci
def test_setup_params(separate_folder):
    from asr.core import read_json
    from asr.setup.scanparams import main
    from pathlib import Path
    params = [3, 4, 5]
    recipe = 'asr.phonons@calculate'
    key = 'kptdensity'
    main(scanparams=[f'{recipe}:{key}'] + list(map(str, params)))

    for i, param in enumerate(params):
        p = Path(f'scanparams{i}', 'params.json')
        assert p.is_file()
        params = read_json(p)
        assert params[recipe][key] == param



from asr.core import command, option, read_json

import numpy as np


def webpanel(row, key_descriptions):
    from asr.database.browser import fig, table

    opt = table(
        row, "Property", ["alphax_lat", "alphay_lat", "alphaz_lat"], key_descriptions
    )

    panel = {
        "title": "Infrared polarizability (RPA)",
        "columns": [[fig("infrax.png"), fig("infraz.png")], [fig("infray.png"), opt]],
        "plot_descriptions": [
            {
                "function": create_plot,
                "filenames": ["infrax.png", "infray.png", "infraz.png"],
            }
        ],
        "sort": 21,
    }

    return [panel]


def create_plot(row, *fnames):
    import matplotlib.pyplot as plt
    from scipy.interpolate import interp1d

    # Get electronic polarizability
    infrareddct = row.data.get("results-asr.infraredpolarizability.json")
    omega_w = infrareddct["omega_w"] * 1e3
    alpha_wvv = infrareddct["alpha_wvv"]

    electrondct = row.data.get("results-asr.polarizability.json")
    alphax_w = electrondct["alphax_w"]
    alphay_w = electrondct["alphay_w"]
    alphaz_w = electrondct["alphaz_w"]
    omegatmp_w = electrondct["frequencies"] * 1e3

    # Get max phonon freq
    phonondata = row.data.get("results-asr.phonons.json")
    maxphononfreq = phonondata.get("omega_kl")[0].max() * 1e3
    maxomega = maxphononfreq * 1.5

    atoms = row.toatoms()
    pbc_c = atoms.pbc
    ndim = int(np.sum(pbc_c))

    realphax = interp1d(omegatmp_w, alphax_w.real)
    imalphax = interp1d(omegatmp_w, alphax_w.imag)
    ax_w = (realphax(omega_w) + 1j * imalphax(omega_w)
            + alpha_wvv[:, 0, 0])
    realphay = interp1d(omegatmp_w, alphay_w.real)
    imalphay = interp1d(omegatmp_w, alphay_w.imag)
    ay_w = (realphay(omega_w) + 1j * imalphay(omega_w)
            + alpha_wvv[:, 1, 1])
    realphaz = interp1d(omegatmp_w, alphaz_w.real)
    imalphaz = interp1d(omegatmp_w, alphaz_w.imag)
    az_w = (realphaz(omega_w) + 1j * imalphaz(omega_w)
            + alpha_wvv[:, 2, 2])

    if ndim == 3:
        epsx_w = 1 + 4 * np.pi * ax_w
        epsy_w = 1 + 4 * np.pi * ay_w
        epsz_w = 1 + 4 * np.pi * az_w
        plt.figure()
        plt.plot(omega_w, epsx_w.imag, label='imag')
        plt.plot(omega_w, epsx_w.real, label='real')
        ax = plt.gca()
        ax.set_title("x-polarization")
        ax.set_xlabel("energy [meV]")
        ax.set_ylabel(r"Dielectric function")
        ax.set_xlim(0, maxomega)
        ax.legend()
        plt.tight_layout()
        plt.savefig(fnames[0])

        plt.figure()
        plt.plot(omega_w, epsy_w.imag, label='imag')
        plt.plot(omega_w, epsy_w.real, label='real')
        ax = plt.gca()
        ax.set_title("y-polarization")
        ax.set_xlabel("energy [meV]")
        ax.set_ylabel(r"Dielectric function")
        ax.set_xlim(0, maxomega)
        ax.legend()
        plt.tight_layout()
        plt.savefig(fnames[1])

        plt.figure()
        plt.plot(omega_w, epsz_w.imag, label='imag')
        plt.plot(omega_w, epsz_w.real, label='real')
        ax = plt.gca()
        ax.set_title("z-polarization")
        ax.set_xlabel("energy [meV]")
        ax.set_ylabel(r"Dielectric function")
        ax.set_xlim(0, maxomega)
        ax.legend()
        plt.tight_layout()
        plt.savefig(fnames[2])
    elif ndim in [2, 1, 0]:
        if ndim == 2:
            unit = r"$\mathrm{\AA}$"
        elif ndim == 1:
            unit = r"$\mathrm{\AA}^2$"
        elif ndim == 0:
            unit = r"$\mathrm{\AA}^3$"
        plt.figure()
        plt.plot(omega_w, ax_w.imag, label='imag')
        plt.plot(omega_w, ax_w.real, label='real')
        ax = plt.gca()
        ax.set_title("x-polarization")
        ax.set_xlabel("energy [meV]")
        ax.set_ylabel(rf"polarizability [{unit}]")
        ax.set_xlim(0, maxomega)
        ax.legend()
        plt.tight_layout()
        plt.savefig(fnames[0])

        plt.figure()
        plt.plot(omega_w, ay_w.imag, label='imag')
        plt.plot(omega_w, ay_w.real, label='real')
        ax = plt.gca()
        ax.set_title("y-polarization")
        ax.set_xlabel("energy [meV]")
        ax.set_ylabel(rf"polarizability [{unit}]")
        ax.set_xlim(0, maxomega)
        ax.legend()
        plt.tight_layout()
        plt.savefig(fnames[1])

        plt.figure()
        plt.plot(omega_w, az_w.imag, label='imag')
        plt.plot(omega_w, az_w.real, label='real')
        ax = plt.gca()
        ax.set_title("z-polarization")
        ax.set_xlabel("energy [meV]")
        ax.set_ylabel(rf"polarizability [{unit}]")
        ax.set_xlim(0, maxomega)
        ax.legend()
        plt.tight_layout()
        plt.savefig(fnames[2])


@command(
    "asr.infraredpolarizability",
    dependencies=["asr.phonons", "asr.borncharges", "asr.polarizability"],
    requires=[
        "structure.json",
        "results-asr.phonons.json",
        "results-asr.borncharges.json",
        "results-asr.polarizability.json",
    ],
    webpanel=webpanel,
)
@option("--nfreq", help="Number of frequency points")
@option("--eta", help="Relaxation rate")
def main(nfreq=300, eta=1e-2):
    from ase.io import read

    # Get relevant atomic structure
    atoms = read("structure.json")

    # Get phonons
    phresults = read_json("results-asr.phonons.json")
    u_ql = phresults["modes_kl"]
    q_qc = phresults["q_qc"]
    omega_ql = phresults["omega_kl"]

    iq_q = np.argwhere((np.abs(q_qc) < 1e-10).all(axis=1))

    assert len(iq_q), "Calculated phonons do not contain Gamma point."

    iq = iq_q[0][0]

    m_a = atoms.get_masses()
    m_inv_x = np.repeat(m_a ** -0.5, 3)
    freqs_l, modes_liv = omega_ql[iq], u_ql[iq]
    modes_xl = modes_liv.reshape(len(freqs_l), -1).T
    modes_xl *= 1 / m_inv_x[:, np.newaxis]

    # Make frequency grid
    fmin = 0
    fmax = omega_ql[0].max() * 3  # Factor of 3 should be enough
    omega_w = np.linspace(fmin, fmax, nfreq)

    # Read born charges
    borndct = read_json("results-asr.borncharges.json")

    # Get other relevant quantities
    m_a = atoms.get_masses()
    cell_cv = atoms.get_cell()
    Z_avv = borndct["Z_avv"]

    # Get phonon polarizability
    alpha_wvv = get_phonon_pol(omega_w, Z_avv, freqs_l, modes_xl, m_a, cell_cv, eta)

    # Normalize according to dimensionality
    pbc_c = atoms.pbc
    if pbc_c.all():
        norm = 1
    else:
        norm = np.abs(np.linalg.det(cell_cv[~pbc_c][:, ~pbc_c]))
    alpha_wvv *= norm

    alphax_lat = alpha_wvv[0, 0, 0].real
    alphay_lat = alpha_wvv[0, 1, 1].real
    alphaz_lat = alpha_wvv[0, 2, 2].real

    elecdict = read_json("results-asr.polarizability.json")
    alphax_el = elecdict["alphax_el"]
    alphay_el = elecdict["alphay_el"]
    alphaz_el = elecdict["alphaz_el"]

    results = {
        "alpha_wvv": alpha_wvv,
        "omega_w": omega_w,
        "alphax_lat": alphax_lat,
        "alphay_lat": alphay_lat,
        "alphaz_lat": alphaz_lat,
        "alphax": alphax_lat + alphax_el,
        "alphay": alphay_lat + alphay_el,
        "alphaz": alphaz_lat + alphaz_el,
    }

    results["__key_descriptions__"] = {
        "alphax_lat": "KVP: Static ionic polarizability, x-direction [Ang]",
        "alphay_lat": "KVP: Static ionic polarizability, y-direction [Ang]",
        "alphaz_lat": "KVP: Static ionic polarizability, z-direction [Ang]",
        "alphax": "KVP: Static total polarizability, x-direction [Ang]",
        "alphay": "KVP: Static total polarizability, y-direction [Ang]",
        "alphaz": "KVP: Static total polarizability, z-direction [Ang]",
    }

    return results


def get_phonon_pol(omega_w, Z_avv, freqs_l, modes_xl, m_a, cell_cv, eta):
    from ase.units import Hartree, Bohr

    Z_vx = Z_avv.swapaxes(0, 1).reshape((3, -1))
    f2_w, D_xw = (freqs_l / Hartree) ** 2, modes_xl

    vol = abs(np.linalg.det(cell_cv)) / Bohr ** 3
    omega_w = omega_w / Hartree
    eta = eta / Hartree
    me = 1822.888
    m_a = m_a * me
    alpha_wvv = np.zeros((len(omega_w), 3, 3), dtype=complex)
    m_x = np.repeat(m_a, 3) ** 0.5
    eta = eta

    for f2, D_x in zip(f2_w, D_xw.T):
        # Neglect acoustic modes
        if f2 < (1e-3 / Hartree) ** 2:
            continue
        DM_x = D_x / m_x
        Z_v = np.dot(Z_vx, DM_x)
        alpha_wvv += (
            np.outer(Z_v, Z_v)[np.newaxis]
            / ((f2 - omega_w ** 2) - 1j * eta * omega_w)[:, np.newaxis, np.newaxis]
        )

    alpha_wvv /= vol
    return alpha_wvv


if __name__ == "__main__":
    main()



import pytest
from .conftest import test_materials


@pytest.mark.ci
@pytest.mark.parametrize("pbc", [[True, ] * 3,
                                 [True, True, False],
                                 [False, False, True]])
def test_setup_strains_get_relevant_strains(separate_folder, pbc):
    from asr.setup.strains import get_relevant_strains

    ij = set(get_relevant_strains(pbc))
    if sum(pbc) == 3:
        ij2 = {(0, 0), (1, 1), (2, 2), (1, 2), (0, 2), (0, 1)}
    elif sum(pbc) == 2:
        ij2 = {(0, 0), (1, 1), (0, 1)}
    elif sum(pbc) == 1:
        ij2 = {(2, 2)}

    assert ij == ij2


@pytest.mark.ci
@pytest.mark.parametrize("atoms", test_materials)
def test_setup_strains(separate_folder, mockgpaw, atoms):
    from asr.setup.strains import (main,
                                   get_strained_folder_name,
                                   get_relevant_strains)
    from asr.core import read_json
    from ase.io import write
    from pathlib import Path
    from asr.setup.params import main as setupparams
    write('structure.json', atoms)
    setupparams(params={'asr.gs@calculate': {
        'calculator': {'mode': 'fd', None: None}}})

    main(strain_percent=1)
    ij = get_relevant_strains(atoms.pbc)
    for i, j in ij:
        name = get_strained_folder_name(1, i, j)
        folder = Path(name)
        assert folder.is_dir()

        paramfile = folder / 'params.json'
        assert paramfile.is_file()

        params = read_json(paramfile)

        assert 'size' in params['asr.relax']['calculator']['kpts']
        assert params['asr.relax']['fixcell']
        assert params['asr.relax']['allow_symmetry_breaking']
        assert params['asr.gs@calculate']['calculator']['mode'] == 'fd'



from .conftest import test_materials, get_webcontent
import pytest


@pytest.mark.parametrize("atoms", test_materials)
def test_projected_bandstructure(separate_folder, mockgpaw, atoms):
    from asr.projected_bandstructure import main
    atoms.write("structure.json")
    main()

    get_webcontent()



import pytest
from ase.db import connect
from ase.build import bulk


@pytest.mark.ci
def test_convex_hull(separate_folder, mockgpaw):
    from ase.calculators.emt import EMT
    from asr.convex_hull import main
    metals = ['Al', 'Cu', 'Ag', 'Au', 'Ni',
              'Pd', 'Pt', 'C']

    energies = {}
    with connect('references.db') as db:
        for uid, element in enumerate(metals):
            atoms = bulk(element)
            atoms.set_calculator(EMT())
            en = atoms.get_potential_energy()
            energies[element] = en
            db.write(atoms, uid=uid)

    db.metadata = {'title': 'Metal references',
                   'legend': 'Metals',
                   'name': '{row.formula}',
                   'link': 'NOLINK',
                   'label': '{row.formula}',
                   'method': 'DFT'}

    metal = 'Ag'
    ag = bulk(metal)
    ag.write('structure.json')
    results = main(databases=['references.db'])
    assert results['hform'] == -energies[metal]



from .conftest import Fe  # , Si
import pytest
from pytest import approx


# XXX How to test for shifted origin, happens for Si
@pytest.mark.ci
@pytest.mark.parametrize("inputatoms", [Fe])
def test_setup_symmetrize(separate_folder, inputatoms):
    import numpy as np
    from asr.setup.symmetrize import main
    from ase.io import read, write
    np.random.seed(1234)
    atoms = inputatoms.copy()
    atoms.rattle(stdev=0.001)
    cell_cv = atoms.get_cell()
    cell_cv += (np.random.rand(3, 3) - 0.5) * 1e-5
    atoms.set_cell(cell_cv)
    write('original.json', atoms)

    main()
    symmetrizedatoms = read('unrelaxed.json')

    assert symmetrizedatoms.cell.cellpar() == approx(inputatoms.cell.cellpar())
    assert (symmetrizedatoms.get_scaled_positions()
            == approx(inputatoms.get_scaled_positions()))
    assert (symmetrizedatoms.get_initial_magnetic_moments()
            == approx(inputatoms.get_initial_magnetic_moments()))



from asr.core import command, argument


tests = [
    {'cli': ['asr run setup.params']},
    {'cli': ['asr run "setup.params asr.relax:ecut 300"'],
     'results': [{'file': 'params.json',
                  'asr.relax:ecut': (250, 0.1)}], 'fails': True},
    {'cli': ['asr run "setup.params :ecut 300"'], 'fails': True},
    {'cli': ['asr run "setup.params asr.relax: 300"'], 'fails': True},
    {'cli': ['asr run "setup.params asr.relax:ecut asr.gs:ecut 300"'],
     'fails': True},
]


@command('asr.setup.params',
         tests=tests)
@argument('params', nargs=-1,
          metavar='recipe:option arg recipe:option arg')
def main(params=None):
    """Compile a params.json file with all options and defaults.

    This recipe compiles a list of all options and their default
    values for all recipes to be used for manually changing values
    for specific options.
    """
    from pathlib import Path
    from asr.core import get_recipes, read_json, write_json
    from ast import literal_eval
    from fnmatch import fnmatch
    import copy
    from asr.core import recursive_update

    defparamdict = {}
    recipes = get_recipes()
    for recipe in recipes:
        defparams = recipe.defparams
        defparamdict[recipe.name] = defparams

    p = Path('params.json')
    if p.is_file():
        paramdict = read_json('params.json')
        # The existing valus in paramdict set the new defaults
        recursive_update(defparamdict, paramdict)
    else:
        paramdict = {}

    if isinstance(params, (list, tuple)):
        # Find recipe:option
        tmpoptions = params[::2]
        tmpargs = params[1::2]
        assert len(tmpoptions) == len(tmpargs), \
            'You must provide a value for each option'
        options = []
        args = []
        for tmpoption, tmparg in zip(tmpoptions, tmpargs):
            assert ':' in tmpoption, 'You have to use the recipe:option syntax'
            recipe, option = tmpoption.split(':')
            if '*' in recipe:
                for tmprecipe in defparamdict:
                    if not fnmatch(tmprecipe, recipe):
                        continue
                    if option in defparamdict[tmprecipe]:
                        options.append(f'{tmprecipe}:{option}')
                        args.append(tmparg)
            else:
                options.append(tmpoption)
                args.append(tmparg)

        for option, value in zip(options, args):
            recipe, option = option.split(':')

            assert option, 'You have to provide an option'
            assert recipe, 'You have to provide a recipe'

            if recipe not in paramdict:
                paramdict[recipe] = {}

            paramtype = type(defparamdict[recipe][option])
            if paramtype == dict:
                value = value.replace('...', 'None:None')
                val = literal_eval(value)
            elif paramtype == bool:
                val = literal_eval(value)
            else:
                val = paramtype(value)
            paramdict[recipe][option] = val
    elif isinstance(params, dict):
        paramdict.update(copy.deepcopy(params))
    else:
        raise NotImplementedError(
            'asr.setup.params is only compatible with'
            f'input lists and dict. Input params: {params}'
        )

    for recipe, options in paramdict.items():
        assert recipe in defparamdict, \
            f'This is an unknown recipe: {recipe}'

        for option, value in options.items():
            assert option in defparamdict[recipe], \
                f'This is an unknown option: {recipe}:{option}'
            if isinstance(value, dict):
                recursive_update(value, defparamdict[recipe][option])
                paramdict[recipe][option] = value

    if paramdict:
        write_json(p, paramdict)
    return paramdict


if __name__ == '__main__':
    main.cli()



from asr.core import command, option


group = 'property'
creates = []  # what files are created


@command('asr.defectformation',
         resources='1:2h',
         dependencies=['asr.setup.defects', 'asr.gs'],
         creates=None)
# later include 'asr.dielectricconstant' and 'asr.relax')
# and also the files that this recipe creates
@option('--pristine', type=str,
        help='Name of the groundstate .gpw file of the pristine system. It '
             'always has to be somewhere within a folder that is called '
             '"pristine" in order to work correctly.')
@option('--defect', type=str,
        help='Name of the groundstate .gpw file of the defect systems. They '
             'always have to be within a folder for the specific defect with '
             'a subfolder calles "charge_q" for the respective chargestate q '
             'in order to work correctly.')
@option('--defect_name',
        help='Runs recipe for all defect folder within your directory when '
             'set to None. Set this option to the name of a desired defect '
             'folder in order for it to run only for this particular defect.')
def main(pristine='gs.gpw', defect='gs.gpw', defect_name=None):
    """
    Calculate formation energy of defects.

    This recipe needs the directory structure that was created with
    setup.defects in order to run properly. It has to be launched within the
    folder of the initial structure, i.e. the folder where setup.defects was
    also executed.
    """
    from ase.io import read
    from asr.core import write_json, read_json
    from gpaw.defects import ElectrostaticCorrections
    from pathlib import Path
    import numpy as np
    q, epsilons, path_gs = check_and_get_general_inputs()
    atoms = read('unrelaxed.json')
    nd = int(np.sum(atoms.get_pbc()))
    gs_dict = read_json('results-asr.gs.json')
    defectformation_dict = {}
    defectformation_dict['gaps_nosoc'] = gs_dict.get('gaps_nosoc')
    defectformation_dict['gaps_soc'] = gs_dict.get('gaps_soc')

    sigma = 2 / (2.0 * np.sqrt(2.0 * np.log(2.0)))
    if nd == 3:
        epsilon = (epsilons[0] + epsilons[1] + epsilons[2]) / 3.
        dim = '3d'
    elif nd == 2:
        epsilon = [(epsilons[0] + epsilons[1]) / 2., epsilons[2]]
        dim = '2d'

    folder_list = []
    p = Path('.')

    # Either run for all defect folders or just a specific one
    if defect_name is None:
        [folder_list.append(x) for x in p.iterdir() if x.is_dir()
            and not x.name == 'pristine_sc']
    else:
        [folder_list.append(x) for x in p.iterdir() if x.is_dir()
            and x.name == defect_name]

    for folder in folder_list:
        s = Path(folder.name)
        sub_folder_list = []
        [sub_folder_list.append(x) for x in s.iterdir() if x.is_dir()]
        e_form = []
        # e_fermi = []
        charges = []
        # e_fermi_calc = []
        for sub_folder in sub_folder_list:
            sub_folder_path = folder.name + '/' + sub_folder.name
            setup_params = read_json(sub_folder_path + '/params.json')
            chargestate = setup_params.get('charge')
            charged_file = find_file_in_folder('gs.gpw',
                                               sub_folder_path)
            elc = ElectrostaticCorrections(pristine=path_gs,
                                           charged=charged_file,
                                           q=chargestate, sigma=sigma,
                                           dimensionality=dim)
            elc.set_epsilons(epsilon)
            if chargestate == 0:
                e_form.append(elc.calculate_uncorrected_formation_energy())
            else:
                e_form.append(elc.calculate_corrected_formation_energy())
            charges.append(chargestate)
            # calc = GPAW(find_file_in_folder('gs.gpw', sub_folder_path))
            # e_fermi_calc.append(calc.get_fermi_level())
            # e_fermi.append(0)
        defectformation_dict[folder.name] = {'formation_energies': e_form,
                                             'chargestates': charges}
        # 'fermi_energies_c': e_fermi_calc
        # 'fermi_energies': e_fermi,
    write_json('defectformation.json', defectformation_dict)

    return None


def check_and_get_general_inputs():
    """Determine whether all necessary files exist.

    Checks if all necessary input files and input parameters for
    this recipe are acessible.
    """
    from asr.core import read_json

    # first, get path of 'gs.gpw' file of pristine_sc, as well as the path of
    # 'dielectricconstant.json' of the pristine system
    path_epsilon = find_file_in_folder('dielectricconstant.json', None)
    path_gs = find_file_in_folder('gs.gpw', 'pristine_sc')
    path_q = find_file_in_folder('general_parameters.json', None)

    # if paths were found correctly, extract epsilon and q
    gen_params = read_json(path_q)
    params_eps = read_json(path_epsilon)
    q = gen_params.get('chargestates')
    epsilons = params_eps.get('local_field')
    if q is not None and epsilons is not None:
        msg = 'INFO: number of chargestates and dielectric constant '
        msg += 'extracted: q = {}, eps = {}'.format(q, epsilons)
        print(msg)
    else:
        msg = 'either number of chargestates and/or dielectric '
        msg += 'constant of the host material could not be extracted'
        raise ValueError(msg)

    if path_gs is not None:
        print('INFO: check of general inputs successful')

    return q, epsilons, path_gs


def find_file_in_folder(filename, foldername):
    """Find a specific file in folder.

    Finds a specific file within a folder starting from your current
    position in the directory tree.
    """
    from pathlib import Path

    p = Path('.')
    find_success = False

    # check in current folder directly if no folder specified
    if foldername is None:
        check_empty = True
        tmp_list = list(p.glob(filename))
        if len(tmp_list) == 1:
            file_path = tmp_list[0]
            print('INFO: found {0}: {1}'.format(filename,
                                                file_path.absolute()))
            find_success = True
        else:
            print('ERROR: no unique {} found in this directory'.format(
                filename))
    else:
        tmp_list = list(p.glob('**/' + foldername))
        check_empty = False
    # check sub_folders
    if len(tmp_list) == 1 and not check_empty:
        file_list = list(p.glob(foldername + '/**/' + filename))
        if len(file_list) == 1:
            file_path = file_list[0]
            print('INFO: found {0} in {1}: {2}'.format(
                filename, foldername, file_path.absolute()))
            find_success = True
        elif len(file_list) == 0:
            print('ERROR: no {} found in this directory'.format(
                filename))
        else:
            print('ERROR: several {0} files in directory tree: {1}'.format(
                filename, tmp_list[0].absolute()))
    elif len(tmp_list) == 0 and not check_empty:
        print('ERROR: no {0} found in this directory tree'.format(
            foldername))
    elif not check_empty:
        print('ERROR: several {0} folders in directory tree: {1}'.format(
            foldername, p.absolute()))

    if not find_success:
        file_path = None

    return str(file_path.absolute())


def collect_data():
    # from ase.io import jsonio
    # from pathlib import Path
    # if not Path('results_defectformation.json').is_file():
    #     return {}, {}, {}

    # kvp = {}
    # data = {}
    # key_descriptions = {}
    # dct = jsonio.decode(Path('results_defectformation.json').read_text())

    return None


# ========================================================================== #
# This function doesn't exist in the new asr version anymore and has to be   #
# changed at a later stage, as well as the way the plots are created (db)    #
# ========================================================================== #
# def postprocessing():
#     from asr.utils import read_json
#
#     formation_dict = read_json('defectformation.json')
#     transitions_dict = {}
#     gap = formation_dict.get('gaps_nosoc').get('gap')
#     for element in formation_dict:
#         if element != 'gaps_soc' and element != 'gaps_nosoc':
#             print(element)
#             plotname = element
#             defect_dict = formation_dict[element]
#             transitions_dict[plotname] = plot_formation_and_transitions(
#                 defect_dict, plotname, gap)
#
#     return transitions_dict


def line_intersection(line1, line2):
    """Get intersection between two lines.

    Helper function to calculate intersection of two given lines.
    """
    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)
    if div == 0:
        raise Exception('lines do not intersect')

    d = (det(*line1), det(*line2))
    x = det(d, xdiff) / div
    y = det(d, ydiff) / div
    return x, y


def line(p1, p2):
    """Define a line between p1 and p2.

    Helper function to define a line.
    """
    A = (p1[1] - p2[1])
    B = (p2[0] - p1[0])
    C = (p1[0] * p2[1] - p2[0] * p1[1])
    return A, B, -C


def intersection(L1, L2):
    """Get intersection between two lines L1 and L2.

    Helper function to calculate intersection of two given lines that were
    defined with the upper 'line' function.
    """
    D = L1[0] * L2[1] - L1[1] * L2[0]
    Dx = L1[2] * L2[1] - L1[1] * L2[2]
    Dy = L1[0] * L2[2] - L1[2] * L2[0]
    if D != 0:
        x = Dx / D
        y = Dy / D
        return x, y
    else:
        return False


def plot_formation_and_transitions(defect_dict, defectname, gap):
    """XXX: Do me.

    Function to plot formation energies versus the Fermi energy and to
    obtain transition points between most stable charge states of a given
    defect
    """
    # from asr.utils import write_json
    import matplotlib.pyplot as plt
    import numpy as np

    # initalize and read formation energies and charge states
    x = []
    y = []
    q = []
    x = np.array(defect_dict['fermi_energies'])
    y = np.array(defect_dict['formation_energies'])
    q = np.array(defect_dict['chargestates'])

    # set general parameters
    x_range = np.array([0, gap])
    x_diff = np.array([[-x[0], x_range[1] - x[0]]])
    y_edges = np.array(
        [[y[0] + x_diff[0][0] * q[0], y[0] + x_diff[0][1] * q[0]]])

    # set general plotting parameters
    plt.figure()
    lw = 1
    linestylelist = ['solid', 'dashdot', 'dashed', 'dotted']
    colorlist = ['black', 'C0', 'C1']
    # plt.ylim(0, max(y_edges[:, 0]) * 1.1)
    plt.ylim(0, 20)
    plt.xlim(x_range[0] - 0.1 * gap, x_range[1] + 0.1 * gap)
    # bbox = {'fc': '0.8', 'pad': 0}

    # initialise np array containing all lines
    linearray = np.array([[line([x_range[0], y_edges[0][0]],
                                [x_range[1], y_edges[0][1]]), q[0]]])

    # initialise plot
    plt.plot(x_range, y_edges[0], color=colorlist[np.sign(q[0])], lw=lw,
             linestyle=linestylelist[abs(q[0])], label='q = {}'.format(q[0]))
    plt.text((2 * x_range[0] - 0.1 * gap) / 2.,
             max(y_edges[:, 0]) * 1.1 / 2., 'valence band',
             {'ha': 'center', 'va': 'center'}, rotation=90)
    plt.text((2 * x_range[1] + 0.1 * gap) / 2.,
             max(y_edges[:, 0]) * 1.1 / 2., 'conduction band',
             {'ha': 'center', 'va': 'center'}, rotation=90)

    # append other lines in a loop
    for i in range(1, len(q)):
        x_diff = np.append(x_diff, [[-x[i], x_range[1] - x[i]]], axis=0)
        y_edges = np.append(y_edges, [[y[i] + x_diff[i][0] * q[i],
                                       y[i] + x_diff[i][1] * q[i]]], axis=0)
        linearray = np.append(linearray, [[line([x_range[0], y_edges[i][0]], [
                              x_range[1], y_edges[i][1]]), q[i]]], axis=0)
        plt.plot(x_range,
                 y_edges[i],
                 color=colorlist[np.sign(q[i])],
                 lw=lw,
                 linestyle=linestylelist[abs(q[i])],
                 label='q = {}'.format(q[i]))

    # flip arrays in order for them to start with positive slope
    linearray_up = np.flip(linearray, axis=0)
    x_diff = np.flip(x_diff, axis=0)
    y_edges = np.flip(y_edges, axis=0)
    q_copy = np.flip(q)

    # IMPORTANT: all of the important arrays have now been reversed!
    # find minimum line at zero energy and create temporary array with lines
    for i in range(len(y_edges)):
        if y_edges[i][0] == min(y_edges[:, 0]):
            # start_index = i
            linearray_up = np.delete(linearray_up, np.s_[0:i], 0)
            trans_array = np.array(
                [[(0, y_edges[i][0]), q_copy[i], q_copy[i]]])
            q_copy = np.delete(q_copy, np.s_[0:i])

    # loop over all lines in linearray_up and calculate intersection points
    while len(linearray_up) > 1:
        linedists = np.array([[intersection(linearray_up[0][0],
                                            linearray_up[1][0]),
                              q_copy[0], q_copy[1]]])
        if len(linearray_up) > 2:
            for j in range(2, len(linearray_up)):
                linedists = np.append(linedists, [[intersection(
                    linearray_up[0][0], linearray_up[j][0]), q_copy[0],
                    q_copy[j]]], axis=0)
        linearray_up = np.delete(linearray_up, 0, 0)
        q_copy = np.delete(q_copy, 0)
        x_list = []
        n = 0
        for element in linedists:
            x_list.append(element[0][0])
            if element[0][0] >= 0 and len(x_list) == 1:
                if element[0][0] == min(x_list):
                    trans_array = np.append(trans_array, [element], axis=0)
                    dropout = n
            elif (element[0][0] >= 0 and element[0][0] == min(x_list)
                  and element[0][0] < trans_array[-1][0][0]):
                trans_array = np.append(trans_array, [element], axis=0)
                trans_array = np.delete(trans_array, -2, 0)
                dropout = n
            n = n + 1
        for i in range(dropout):
            linearray_up = np.delete(linearray_up, np.s_[0:i], 0)
    for i in range(len(y_edges)):
        if y_edges[i][1] == min(y_edges[:, 1]):
            trans_array = np.append(trans_array, [[(x_range[1], y_edges[i][1]),
                                                   trans_array[-1][2],
                                                   trans_array[-1][2]]],
                                    axis=0)

    # plot the results and save the figure
    for element in trans_array:
        ratio = element[0][1] / max(y_edges[:, 0])
        plt.axvline(x=element[0][0], ymax=ratio, color='C3',
                    linestyle=(0, (5, 10)))
    plt.axvspan(x_range[0] - 0.1 * gap, x_range[0], color='lightgrey')
    plt.axvspan(x_range[1] + 0.1 * gap, x_range[1], color='lightgrey')
    x_val = [x[0] for x in trans_array[:, 0]]
    y_val = [x[1] for x in trans_array[:, 0]]
    # plt.plot(x_val, y_val, marker='D', linestyle='-', color='C3')
    plt.plot(x_val, y_val, marker='D', linestyle='', color='C3')
    plt.xlabel(r'$E_{F}$ in eV')
    plt.ylabel(r'$E_{formation}$ in eV')
    plotname = 'plot_{}.png'.format(defectname)
    plt.legend()
    plt.savefig(plotname)

    return trans_array


# def webpanel(row, key_descriptions):
#    from asr.database.browser import fig, table
#
#    if 'something' not in row.data:
#        return None, []
#
#    table1 = table(row,
#                   'Property',
#                   ['something'],
#                   kd=key_descriptions)
#    panel = ('Title',
#             [[fig('something.png'), table1]])
#    things = [(create_plot, ['something.png'])]
#    return panel, things


# def create_plot(row, fname):
#    import matplotlib.pyplot as plt
#
#    data = row.data.something
#    fig = plt.figure()
#    ax = fig.gca()
#    ax.plot(data.things)
#    plt.savefig(fname)


if __name__ == '__main__':
    main.cli()



import pytest
from .conftest import test_materials, get_webcontent


@pytest.mark.ci
@pytest.mark.parametrize("atoms", test_materials)
def test_polarizability(separate_folder, mockgpaw, atoms):
    from asr.polarizability import main
    atoms.write('structure.json')
    main()
    content = get_webcontent()
    assert "polarizability" in content, content



from asr.core import command, option
import numpy as np


def webpanel(row, key_descriptions):
    data = row.data.get('results-asr.deformationpotentials.json')

    defpot = data['deformation_potentials']
    vbmdef = (defpot[0, 0] + defpot[1, 0]) / 2
    cbmdef = (defpot[0, 1] + defpot[1, 1]) / 2
    rows = [['Uniaxial deformation potential at VBM', f'{cbmdef:0.2f} eV'],
            ['Uniaxial deformation potential at CBM', f'{vbmdef:0.2f} eV']]
    panel = {'title': 'Basic electronic properties (PBE)',
             'columns': [[{'type': 'table',
                           'header': ['Property', ''],
                           'rows': rows}]],
             'sort': 11}
    return [panel]


@command(webpanel=webpanel)
@option('--strains', help='Strain percentages')
@option('--ktol', help='Distance that ')
def main(strains=[-1.0, 0.0, 1.0], ktol=0.1):
    """Calculate deformation potentials.

    Calculate the deformation potential both with and without spin orbit
    coupling, for both the conduction band and the valence band, and return as
    a dictionary.
    """
    from asr.setup.strains import (get_strained_folder_name,
                                   get_relevant_strains)
    from asr.core import read_json
    from ase.io import read
    strains = sorted(strains)
    atoms = read('structure.json')
    ij = get_relevant_strains(atoms.pbc)

    ij_to_voigt = [[0, 5, 4],
                   [5, 1, 3],
                   [4, 3, 2]]

    # Edges have dimension (3, 6, 2) =
    # (#strains_percentages, #strains, (vbm, cbm))
    # Because np.polyfit likes that
    edges_pin = np.zeros((3, 6, 2), float)
    edges_nosoc_pin = np.zeros((3, 6, 2), float)

    gsresults = read_json('results-asr.gs.json')

    k0_vbm_c = gsresults['k_vbm_c']
    k0_cbm_c = gsresults['k_cbm_c']

    for i, j in ij:
        for ip, strain in enumerate(strains):
            folder = get_strained_folder_name(strain, i, j)
            gsresults = read_json(folder / 'results-asr.gs.json')
            k_vbm_c = gsresults['k_vbm_c']
            k_cbm_c = gsresults['k_cbm_c']
            difference = k_vbm_c - k0_vbm_c
            difference -= np.round(difference)
            assert (np.abs(difference) < ktol).all(), \
                (f'{folder}: i={i} j={j} strain={strain}: VBM has '
                 f'changed location in reciprocal space upon straining. '
                 f'{k0_vbm_c} -> {k_vbm_c} (Delta_c={difference})')
            difference = k_cbm_c - k0_cbm_c
            difference -= np.round(difference)
            assert (np.abs(difference) < ktol).all(), \
                (f'{folder}: i={i} j={j} strain={strain}: CBM has '
                 f'changed location in reciprocal space upon straining. '
                 f'{k0_cbm_c} -> {k_cbm_c} (Delta_c={difference})')
            evac = gsresults['evac']
            edges_pin[ip, ij_to_voigt[i][j], 0] = gsresults['vbm'] - evac
            edges_nosoc_pin[ip, ij_to_voigt[i][j], 0] = \
                gsresults['gaps_nosoc']['vbm'] - evac
            edges_pin[ip, ij_to_voigt[i][j], 1] = gsresults['cbm'] - evac
            edges_nosoc_pin[ip, ij_to_voigt[i][j], 1] = \
                gsresults['gaps_nosoc']['cbm'] - evac

    results = {'edges': edges_pin,
               'edges_nosoc': edges_nosoc_pin}

    for soc in (True, False):
        if soc:
            edges_pin = edges_pin
        else:
            edges_pin = edges_nosoc_pin

        deformation_potentials = np.zeros(np.shape(edges_pin)[1:])
        for idx, band_edge in enumerate(['vbm', 'cbm']):
            D = np.polyfit(strains, edges_pin[:, :, idx], 1)[0] * 100
            deformation_potentials[:, idx] = D
        results[['deformation_potentials_nosoc',
                 'deformation_potentials'][soc]] = \
            deformation_potentials.tolist()

    return results


if __name__ == '__main__':
    main.cli()



from asr.core import command, argument, option
import numpy as np
from pathlib import Path


def apply_substitution(atoms, substitution):
    from ase.data import covalent_radii
    new_atoms = atoms.copy()
    new_numbers = [substitution[number] for number in atoms.numbers]

    # Scale in-plane lattice vectors by the harmonic mean of the covalent_radii
    sf = (np.product([covalent_radii[n] for n in new_numbers])
          / np.product([covalent_radii[n] for n in atoms.numbers]))
    sf = pow(sf, 1. / len(new_numbers))
    sf = [sf if atoms.pbc[n] else 1 for n in range(3)]
    newcell = np.diag(sf).dot(new_atoms.get_cell())
    new_atoms.set_cell(newcell, scale_atoms=True)

    # update the atomic numbers
    new_atoms.numbers = new_numbers
    return new_atoms


def find_substitutions(number, data, threshold):
    from ase.data import atomic_numbers
    row = data[number]
    substitutions = np.where(row > threshold)[0]
    allowed_elements = [
        'H', 'He',
        'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne',
        'Na', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar',

        'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn',
        'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr',

        'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd',
        'In', 'Sn', 'Sb', 'Te', 'I', 'Xe',
        # 6
        'Cs', 'Ba', 'La', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg',
        'Tl', 'Pb', 'Bi', 'Rn']
    allowed_numbers = [atomic_numbers[e] for e in allowed_elements]
    substitutions = [s for s in substitutions if s in allowed_numbers]
    return substitutions


def generate_structures(prototype, p_ab, threshold=0.08):
    from itertools import product
    numbers = list(set(prototype.numbers))
    substitutions = [find_substitutions(number, p_ab, threshold)
                     for number in numbers]
    for substitution in product(*substitutions):
        new_numbers = list(set(substitution))
        if len(new_numbers) != len(numbers):
            continue
        yield prototype, dict(zip(numbers, substitution))


def get_p_ab():
    """Get similarity matrix.

    The data is saved as a matrix of counts, so that s_ab gives the number of
    times that a can substitute for b in the icsd. This is be normalized,
    to give a probability of succesful substitution. See

    """
    name = Path(__file__).parent / 'substitution.dat'
    s_ab = np.loadtxt(name)
    from ase.utils import seterr
    with seterr(divide='ignore', invalid='ignore'):
        tmp = s_ab ** 2 / s_ab.sum(axis=0)[None, :] / s_ab.sum(axis=1)[:, None]
    tmp[np.isnan(tmp)] = 0
    np.fill_diagonal(tmp, 1)

    return np.sqrt(tmp)


@command('asr.setup.decorate')
@argument('atoms', type=str)
@option('--threshold',
        help='Threshold of likelyhood of two atomic species to subsititute')
@option('--database')
def main(atoms, threshold=0.08, database='decorated.db'):
    """Create similar atomic structures.

    This recipe can substitute atoms in an atomic structure with other similar
    atoms. In this case, similarity is defined as a probability describing the
    number of experimentally known atomic structures which only differ
    by a simple substitution, say Si -> Ge.

    The number of coexisting atomic structures has been analyzed in Ref. XXX
    and this recipe is converting this number to a probability.

    The threshold option limits the number of performed atomic substitions to
    the ones that have a probability larger than the threshold.

    By default the decorated atomic structures will be packed into an ASE
    database which can be unpacked into a folder structure using the
    "setup.unpackdatabase" recipe.

    Examples
    --------
    Perform likely substitions of atomic structure in structure.json
        asr run "setup.decorate structure.json"
    """
    from ase.db import connect
    from ase.io import read
    from ase.data import chemical_symbols
    p_ab = get_p_ab()
    db = connect(database)
    atoms = read(atoms)
    for structure, subs in generate_structures(atoms, p_ab,
                                               threshold=threshold):
        structure = apply_substitution(structure, subs)
        formula = structure.symbols.formula
        explanation = ''
        for i, j in subs.items():
            a = chemical_symbols[i]
            b = chemical_symbols[j]
            prob = p_ab[i, j]
            explanation += f'{a}->{b} (P={prob:.3f})'
        print(f'Created {formula:metal}: {explanation}')
        db.write(structure)


if __name__ == "__main__":
    main.cli()
